# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import errno
import hashlib
import logging
import os
import pathlib
import re
import shutil
import stat
import subprocess
import sys
from contextlib import contextmanager, suppress
from typing import Callable, Generator, List, Optional, Pattern, Set

from snapcraft.internal import common, errors

if sys.version_info < (3, 6):
    import sha3  # noqa


logger = logging.getLogger(__name__)


def replace_in_file(
    directory: str, file_pattern: Pattern, search_pattern: Pattern, replacement: str
) -> None:
    """Searches and replaces patterns that match a file pattern.

    :param str directory: The directory to look for files.
    :param str file_pattern: The file pattern to match inside directory.
    :param search_pattern: A re.compile'd pattern to search for within
                           matching files.
    :param str replacement: The string to replace the matching search_pattern
                            with.
    """

    for root, directories, files in os.walk(directory):
        for file_name in files:
            if file_pattern.match(file_name):
                file_path = os.path.join(root, file_name)
                # Don't bother trying to rewrite a symlink. It's either invalid
                # or the linked file will be rewritten on its own.
                if not os.path.islink(file_path):
                    search_and_replace_contents(file_path, search_pattern, replacement)


def search_and_replace_contents(
    file_path: str, search_pattern: Pattern, replacement: str
) -> None:
    """Search file and replace any occurrence of pattern with replacement.

    :param str file_path: Path of file to be searched.
    :param re.RegexObject search_pattern: Pattern for which to search.
    :param str replacement: The string to replace pattern.
    """
    try:
        with open(file_path, "r+") as f:
            try:
                original = f.read()
            except UnicodeDecodeError:
                # This was probably a binary file. Skip it.
                return

            replaced = search_pattern.sub(replacement, original)
            if replaced != original:
                f.seek(0)
                f.truncate()
                f.write(replaced)
    except PermissionError as e:
        logger.warning(
            "Unable to open {path} for writing: {error}".format(path=file_path, error=e)
        )


def link_or_copy(source: str, destination: str, follow_symlinks: bool = False) -> None:
    """Hard-link source and destination files. Copy if it fails to link.

    Hard-linking may fail (e.g. a cross-device link, or permission denied), so
    as a backup plan we just copy it.

    :param str source: The source to which destination will be linked.
    :param str destination: The destination to be linked to source.
    :param bool follow_symlinks: Whether or not symlinks should be followed.
    """

    try:
        if not follow_symlinks and os.path.islink(source):
            copy(source, destination)
        else:
            link(source, destination, follow_symlinks=follow_symlinks)
    except OSError as e:
        if e.errno == errno.EEXIST and not os.path.isdir(destination):
            # os.link will fail if the destination already exists, so let's
            # remove it and try again.
            os.remove(destination)
            link_or_copy(source, destination, follow_symlinks)
        else:
            copy(source, destination, follow_symlinks=follow_symlinks)


def link(source: str, destination: str, *, follow_symlinks: bool = False) -> None:
    """Hard-link source and destination files.

    :param str source: The source to which destination will be linked.
    :param str destination: The destination to be linked to source.
    :param bool follow_symlinks: Whether or not symlinks should be followed.

    :raises SnapcraftCopyFileNotFoundError: If source doesn't exist.
    """
    # Note that follow_symlinks doesn't seem to work for os.link, so we'll
    # implement this logic ourselves using realpath.
    source_path = source
    if follow_symlinks:
        source_path = os.path.realpath(source)

    if not os.path.exists(os.path.dirname(destination)):
        create_similar_directory(
            os.path.dirname(source_path), os.path.dirname(destination)
        )
    # Setting follow_symlinks=False in case this bug is ever fixed
    # upstream-- we want this function to continue supporting NOT following
    # symlinks.
    try:
        os.link(source_path, destination, follow_symlinks=False)
    except FileNotFoundError:
        raise errors.SnapcraftCopyFileNotFoundError(source)


def copy(source: str, destination: str, *, follow_symlinks: bool = False) -> None:
    """Copy source and destination files.

    This function overwrites the destination if it already exists, and also
    tries to copy ownership information.

    :param str source: The source to be copied to destination.
    :param str destination: Where to put the copy.
    :param bool follow_symlinks: Whether or not symlinks should be followed.

    :raises SnapcraftCopyFileNotFoundError: If source doesn't exist.
    """
    # If os.link raised an I/O error, it may have left a file behind. Skip on
    # OSError in case it doesn't exist or is a directory.
    with suppress(OSError):
        os.unlink(destination)

    try:
        shutil.copy2(source, destination, follow_symlinks=follow_symlinks)
    except FileNotFoundError:
        raise errors.SnapcraftCopyFileNotFoundError(source)
    uid = os.stat(source, follow_symlinks=follow_symlinks).st_uid
    gid = os.stat(source, follow_symlinks=follow_symlinks).st_gid
    try:
        os.chown(destination, uid, gid, follow_symlinks=follow_symlinks)
    except PermissionError as e:
        logger.debug(
            "Unable to chown {destination}: {error}".format(
                destination=destination, error=e
            )
        )


def link_or_copy_tree(
    source_tree: str,
    destination_tree: str,
    ignore: Callable[[str, List[str]], List[str]] = None,
    copy_function: Callable[..., None] = link_or_copy,
) -> None:
    """Copy a source tree into a destination, hard-linking if possible.

    :param str source_tree: Source directory to be copied.
    :param str destination_tree: Destination directory. If this directory
                                 already exists, the files in `source_tree`
                                 will take precedence.
    :param callable ignore: If given, called with two params, source dir and
                            dir contents, for every dir copied. Should return
                            list of contents to NOT copy.
    :param callable copy_function: Callable that actually copies.
    """

    if not os.path.isdir(source_tree):
        raise errors.SnapcraftEnvironmentError(
            "{!r} is not a directory".format(source_tree)
        )

    if not os.path.isdir(destination_tree) and (
        os.path.exists(destination_tree) or os.path.islink(destination_tree)
    ):
        raise errors.SnapcraftEnvironmentError(
            "Cannot overwrite non-directory {!r} with directory "
            "{!r}".format(destination_tree, source_tree)
        )

    create_similar_directory(source_tree, destination_tree)

    destination_basename = os.path.basename(destination_tree)

    for root, directories, files in os.walk(source_tree, topdown=True):
        ignored: Set[str] = set()
        if ignore is not None:
            ignored = set(ignore(root, directories + files))

        # Don't recurse into destination tree if it's a subdirectory of the
        # source tree.
        if os.path.relpath(destination_tree, root) == destination_basename:
            ignored.add(destination_basename)

        if ignored:
            # Prune our search appropriately given an ignore list, i.e. don't
            # walk into directories that are ignored.
            directories[:] = [d for d in directories if d not in ignored]

        for directory in directories:
            source = os.path.join(root, directory)
            # os.walk doesn't by default follow symlinks (which is good), but
            # it includes symlinks that are pointing to directories in the
            # directories list. We want to treat it as a file, here.
            if os.path.islink(source):
                files.append(directory)
                continue

            destination = os.path.join(
                destination_tree, os.path.relpath(source, source_tree)
            )

            create_similar_directory(source, destination)

        for file_name in set(files) - ignored:
            source = os.path.join(root, file_name)
            destination = os.path.join(
                destination_tree, os.path.relpath(source, source_tree)
            )

            copy_function(source, destination)


def create_similar_directory(source: str, destination: str) -> None:
    """Create a directory with the same permission bits and owner information.

    :param str source: Directory from which to copy name, permission bits, and
                       owner information.
    :param str destintion: Directory to create and to which the `source`
                           information will be copied.
    """

    stat = os.stat(source, follow_symlinks=False)
    uid = stat.st_uid
    gid = stat.st_gid
    os.makedirs(destination, exist_ok=True)

    # Windows does not have "os.chown" implementation and copystat
    # is unlikely to be useful, so just bail after creating directory.
    if sys.platform == "win32":
        return

    try:
        os.chown(destination, uid, gid, follow_symlinks=False)
    except PermissionError as exception:
        logger.debug("Unable to chown {}: {}".format(destination, exception))

    shutil.copystat(source, destination, follow_symlinks=False)


@contextmanager
def requires_command_success(
    command: str, not_found_fmt: str = None, failure_fmt: str = None
) -> Generator:
    if isinstance(command, str):
        cmd_list = command.split()
    else:
        raise TypeError("command must be a string.")
    kwargs = dict(command=command, cmd_list=cmd_list)
    try:
        subprocess.check_call(cmd_list, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    except FileNotFoundError:
        if not_found_fmt is not None:
            kwargs["fmt"] = not_found_fmt
        raise errors.RequiredCommandNotFound(**kwargs)
    except subprocess.CalledProcessError:
        if failure_fmt is not None:
            kwargs["fmt"] = failure_fmt
        raise errors.RequiredCommandFailure(**kwargs)
    yield


@contextmanager
def requires_path_exists(path: str, error_fmt: str = None) -> Generator:
    if not os.path.exists(path):
        kwargs = dict(path=path)
        if error_fmt is not None:
            kwargs["fmt"] = error_fmt
        raise errors.RequiredPathDoesNotExist(**kwargs)
    yield


def _file_reader_iter(path: str, block_size=2 ** 20):
    with open(path, "rb") as f:
        block = f.read(block_size)
        while len(block) > 0:
            yield block
            block = f.read(block_size)


def calculate_sha3_384(path: str) -> str:
    """Calculate sha3 384 hash, reading the file in 1MB chunks."""
    return calculate_hash(path, algorithm="sha3_384")


def calculate_hash(path: str, *, algorithm: str) -> str:
    """Calculate the hash for path with algorithm."""
    # This will raise an AttributeError if algorithm is unsupported
    hasher = getattr(hashlib, algorithm)()

    for block in _file_reader_iter(path):
        hasher.update(block)
    return hasher.hexdigest()


def get_host_tool_path(*, command_name: str, package_name: str) -> pathlib.Path:
    """Return the path of command_name found on the host.

    Uses shutil.which() to find the host-provided command.

    :param command_name: name of the command to resolve a path for.
    :param package_name: package <command_name> is usually found in.
    :raises SnapcraftHostToolMissingError: if command_name not found.
    :return: Path to command.
    """
    command_path = shutil.which(command_name)
    if command_path is None:
        raise errors.SnapcraftHostToolNotFoundError(
            command_name=command_name, package_name=package_name
        )

    return pathlib.Path(command_path)


def get_snap_tool_path(command_name: str) -> str:
    """Return the path command found in the snap.

    If snapcraft is not running as a snap, shutil.which() is used
    to resolve the command using PATH.

    :param command_name: the name of the command to resolve a path for.
    :raises ToolMissingError: if command_name was not found.
    :return: Path to command
    """
    if common.is_snap():
        snap_path = os.getenv("SNAP")
        if snap_path is None:
            raise RuntimeError("SNAP not defined, but SNAP_NAME is?")

        command_path = _find_command_path_in_root(snap_path, command_name)
    else:
        command_path = shutil.which(command_name)

    if command_path is None:
        raise errors.ToolMissingError(command_name=command_name)

    return command_path


def _find_command_path_in_root(root, command_name: str) -> Optional[str]:
    for bin_directory in (
        os.path.join("usr", "local", "sbin"),
        os.path.join("usr", "local", "bin"),
        os.path.join("usr", "sbin"),
        os.path.join("usr", "bin"),
        os.path.join("sbin"),
        os.path.join("bin"),
    ):
        path = os.path.join(root, bin_directory, command_name)
        if os.path.exists(path):
            return path

    return None


def get_linker_version_from_file(linker_file: str) -> str:
    """Returns the version of the linker from linker_file.

    linker_file must be of the format ld-<linker_version>.so

    :param str linker_file: a valid file path or basename representing
                            the linker from libc6 or related.
    :returns: the version extracted from the linker file.
    :rtype: string
    :raises snapcraft.internal.errors.errors.SnapcraftEnvironmentError:
       if linker_file is not of the expected format.
    """
    m = re.search(r"ld-(?P<linker_version>[\d.]+).so$", linker_file)
    if not m:
        # This is a programmatic error, we don't want to be friendly
        # about this.
        raise errors.SnapcraftEnvironmentError(
            "The format for the linker should be of the of the form "
            "<root>/ld-<X>.<Y>.so. {!r} does not match that format. "
            "Ensure you are targeting an appropriate base".format(linker_file)
        )
    linker_version = m.group("linker_version")

    return linker_version


def get_resolved_relative_path(relative_path: str, base_directory: str) -> str:
    """Resolve path components against target base_directory.

    If the resulting target path is a symlink, it will not be followed.
    Only the path's parents are fully resolved against base_directory,
    and the relative path is returned.

    :param str relative_path: Path of target, relative to base_directory.
    :param str base_directory: Base path of target.
    :return: Resolved path, relative to base_directory.
    :rtype: str
    """
    parent_relpath, filename = os.path.split(relative_path)
    parent_abspath = os.path.realpath(os.path.join(base_directory, parent_relpath))

    filename_abspath = os.path.join(parent_abspath, filename)
    filename_relpath = os.path.relpath(filename_abspath, base_directory)

    return filename_relpath


def _remove_readonly(func, path, excinfo):
    # Try setting file to writeable if error occurs during rmtree.
    # Known to be required on Windows where file is not "writeable",
    # but it is owned by the user (who can set file permissions).
    os.chmod(path, stat.S_IWRITE)
    func(path)


def rmtree(path: str) -> None:
    """Cross-platform rmtree implementation."""
    shutil.rmtree(path, onerror=_remove_readonly)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import http.client
import http.server
import json
import os
import tempfile
import urllib.parse

import fixtures
from testtools.matchers import Equals

from tests import fixture_setup, unit


class TempCWDTestCase(unit.TestCase):
    def test_with_TEMPDIR_env_var(self):
        with tempfile.TemporaryDirectory() as test_tmp_dir:
            with fixtures.EnvironmentVariable("TMPDIR", test_tmp_dir):
                temp_cwd_fixture = fixture_setup.TempCWD()
                self.useFixture(temp_cwd_fixture)

        self.assertThat(os.path.dirname(temp_cwd_fixture.path), Equals(test_tmp_dir))


class TestFakeServer(http.server.HTTPServer):
    def __init__(self, server_address):
        super().__init__(server_address, TestFakeRequestHandler)


class TestFakeRequestHandler(http.server.BaseHTTPRequestHandler):
    def do_GET(self):
        self.send_response(200)
        self.send_header("Content-Type", "application/json")
        self.end_headers()
        self.wfile.write(json.dumps("{}").encode())

    def log_message(*args):
        # Do not print anything during the tests.
        pass


class FakeServerRunningTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()
        self.fake_server_fixture = fixture_setup.FakeServerRunning()
        self.fake_server_fixture.fake_server = TestFakeServer

    def start_fake_server(self):
        self.useFixture(self.fake_server_fixture)
        self.netloc = urllib.parse.urlparse(self.fake_server_fixture.url).netloc

    def do_request(self, method, path):
        connection = http.client.HTTPConnection(self.netloc)
        self.addCleanup(connection.close)
        connection.request(method, path)
        response = connection.getresponse()
        return response.status

    def assert_server_not_running(self):
        self.assertRaises(Exception, self.do_request, "GET", "/")

    def test_server_must_start_and_stop(self):
        self.addCleanup(self.assert_server_not_running)
        self.start_fake_server()
        status = self.do_request("GET", "/")
        self.assertThat(status, Equals(200))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from snapcraft.internal.project_loader._extensions.kde_neon import ExtensionImpl


def test_extension_core18():
    kde_neon_extension = ExtensionImpl(
        extension_name="kde-neon", yaml_data=dict(base="core18")
    )

    assert kde_neon_extension.root_snippet == {
        "assumes": ["snapd2.43"],
        "plugs": {
            "desktop": {"mount-host-font-cache": False},
            "icon-themes": {
                "interface": "content",
                "target": "$SNAP/data-dir/icons",
                "default-provider": "gtk-common-themes",
            },
            "sound-themes": {
                "interface": "content",
                "target": "$SNAP/data-dir/sounds",
                "default-provider": "gtk-common-themes",
            },
            "kde-frameworks-5-plug": {
                "content": "kde-frameworks-5-core18-all",
                "interface": "content",
                "target": "$SNAP/kf5",
                "default-provider": "kde-frameworks-5-core18",
            },
        },
        "environment": {"SNAP_DESKTOP_RUNTIME": "$SNAP/kf5"},
        "hooks": {
            "configure": {
                "plugs": ["desktop"],
                "command-chain": ["snap/command-chain/hooks-configure-desktop"],
            }
        },
    }
    assert kde_neon_extension.app_snippet == {
        "command-chain": ["snap/command-chain/desktop-launch"],
        "plugs": ["desktop", "desktop-legacy", "opengl", "wayland", "x11"],
    }
    assert kde_neon_extension.part_snippet == dict()
    assert kde_neon_extension.parts == {
        "kde-neon-extension": {
            "source": "$SNAPCRAFT_EXTENSIONS_DIR/desktop",
            "source-subdir": "kde-neon",
            "plugin": "make",
            "make-parameters": ["PLATFORM_PLUG=kde-frameworks-5-plug"],
            "build-packages": ["g++"],
            "build-snaps": ["kde-frameworks-5-core18-sdk/latest/stable"],
        }
    }


def test_extension_core20():
    kde_neon_extension = ExtensionImpl(
        extension_name="kde-neon", yaml_data=dict(base="core20")
    )

    assert kde_neon_extension.root_snippet == {
        "assumes": ["snapd2.43"],
        "environment": {"SNAP_DESKTOP_RUNTIME": "$SNAP/kf5"},
        "hooks": {
            "configure": {
                "command-chain": ["snap/command-chain/hooks-configure-desktop"],
                "plugs": ["desktop"],
            }
        },
        "plugs": {
            "desktop": {"mount-host-font-cache": False},
            "icon-themes": {
                "default-provider": "gtk-common-themes",
                "interface": "content",
                "target": "$SNAP/data-dir/icons",
            },
            "kde-frameworks-5-plug": {
                "content": "kde-frameworks-5-qt-5-15-core20-all",
                "default-provider": "kde-frameworks-5-qt-5-15-core20",
                "interface": "content",
                "target": "$SNAP/kf5",
            },
            "sound-themes": {
                "default-provider": "gtk-common-themes",
                "interface": "content",
                "target": "$SNAP/data-dir/sounds",
            },
        },
    }
    assert kde_neon_extension.app_snippet == {
        "command-chain": ["snap/command-chain/desktop-launch"],
        "plugs": ["desktop", "desktop-legacy", "opengl", "wayland", "x11"],
    }
    assert kde_neon_extension.part_snippet == {
        "build-environment": [
            {
                "SNAPCRAFT_CMAKE_ARGS": "-DCMAKE_FIND_ROOT_PATH=/snap/kde-frameworks-5-qt-5-15-core20-sdk/current"
            }
        ]
    }
    assert kde_neon_extension.parts == {
        "kde-neon-extension": {
            "build-packages": ["g++"],
            "build-snaps": ["kde-frameworks-5-qt-5-15-core20-sdk/latest/candidate"],
            "make-parameters": ["PLATFORM_PLUG=kde-frameworks-5-plug"],
            "plugin": "make",
            "source": "$SNAPCRAFT_EXTENSIONS_DIR/desktop",
            "source-subdir": "kde-neon",
        }
    }


def test_supported_bases():
    assert ExtensionImpl.get_supported_bases() == ("core18", "core20")


def test_supported_confinement():
    assert ExtensionImpl.get_supported_confinement() == ("strict", "devmode")


def test_experimental_core20():
    kde_neon_extension = ExtensionImpl(
        extension_name="kde-neon", yaml_data=dict(base="core20")
    )
    assert kde_neon_extension.is_experimental(base="core20") is True


def test_experimental_core18():
    kde_neon_extension = ExtensionImpl(
        extension_name="kde-neon", yaml_data=dict(base="core18")
    )
    assert kde_neon_extension.is_experimental(base="core18") is False



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The colcon plugin for ROS2 parts.

    - colcon-packages:
      (list of strings)
      List of colcon packages to build. If not specified, all packages in the
      workspace will be built. If set to an empty list ([]), no packages will
      be built, which could be useful if you only want ROS debs in the snap.

    - colcon-packages-ignore:
      (list of strings)
      List of colcon packages to ignore. If not specified or set to an empty
      list ([]), no packages will be ignored.

    - colcon-cmake-args:
      (list of strings)
      Arguments to pass to cmake projects. Note that any arguments here which match
      colcon arguments need to be prefixed with a space. This can be done by quoting
      each argument with a leading space.

    - colcon-catkin-cmake-args:
      (list of strings)
      Arguments to pass to catkin packages. Note that any arguments here which match
      colcon arguments need to be prefixed with a space. This can be done by quoting
      each argument with a leading space.

    - colcon-ament-cmake-args:
      (list of strings)
      Arguments to pass to ament_cmake packages. Note that any arguments here which
      match colcon arguments need to be prefixed with a space. This can be done by
      quoting each argument with a leading space.

    This plugin expects the build-environment `ROS_DISTRO` and `ROS_BUILD_BASE`
    to be populated by the `ros2-<distro>` extension.

    This plugin also expects certain variables that are specified by the extension,
    specific to the ROS distro.  If not using the extension, set these in your
    `build-environment`:
      - ROS_DISTRO: "foxy"
"""

from typing import Any, Dict, List, Set

from snapcraft.plugins.v2 import _ros


class ColconPlugin(_ros.RosPlugin):
    @classmethod
    def get_schema(cls) -> Dict[str, Any]:
        return {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "type": "object",
            "additionalProperties": False,
            "properties": {
                "colcon-ament-cmake-args": {
                    "type": "array",
                    "minItems": 0,
                    "items": {"type": "string"},
                    "default": [],
                },
                "colcon-catkin-cmake-args": {
                    "type": "array",
                    "minItems": 0,
                    "items": {"type": "string"},
                    "default": [],
                },
                "colcon-cmake-args": {
                    "type": "array",
                    "minItems": 0,
                    "items": {"type": "string"},
                    "default": [],
                },
                "colcon-packages": {
                    "type": "array",
                    "minItems": 0,
                    "uniqueItems": True,
                    "items": {"type": "string"},
                },
                "colcon-packages-ignore": {
                    "type": "array",
                    "minItems": 0,
                    "uniqueItems": True,
                    "items": {"type": "string"},
                    "default": [],
                },
            },
        }

    def get_build_packages(self) -> Set[str]:
        return super().get_build_packages() | {
            "python3-colcon-common-extensions",
            "python3-rosinstall",
            "python3-wstool",
        }

    def get_build_environment(self) -> Dict[str, str]:
        env = super().get_build_environment()
        env.update(
            {
                "AMENT_PYTHON_EXECUTABLE": "/usr/bin/python3",
                "COLCON_PYTHON_EXECUTABLE": "/usr/bin/python3",
            }
        )

        return env

    def _get_build_commands(self) -> List[str]:
        cmd = [
            "colcon",
            "build",
            "--base-paths",
            "$SNAPCRAFT_PART_SRC",
            "--build-base",
            "$SNAPCRAFT_PART_BUILD",
            "--merge-install",
            "--install-base",
            "$SNAPCRAFT_PART_INSTALL",
        ]

        if self.options.colcon_packages_ignore:
            cmd.extend(["--packages-ignore", *self.options.colcon_packages_ignore])

        if self.options.colcon_packages:
            cmd.extend(["--packages-select", *self.options.colcon_packages])

        if self.options.colcon_ament_cmake_args:
            cmd.extend(["--ament-cmake-args", *self.options.colcon_ament_cmake_args])

        if self.options.colcon_catkin_cmake_args:
            cmd.extend(["--catkin-cmake-args", *self.options.colcon_catkin_cmake_args])

        # Specify the number of workers
        cmd.extend(["--parallel-workers", "${SNAPCRAFT_PARALLEL_BUILD_COUNT}"])

        return [" ".join(cmd)]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import absolute_import, unicode_literals

# FIXME: snapcraft targets the '16' series, hardcode it until more choices
# become available server side -- vila 2016-04-22
DEFAULT_SERIES = "16"
SCAN_STATUS_POLL_DELAY = 5
SCAN_STATUS_POLL_RETRIES = 5

STORE_DASHBOARD_URL = "https://dashboard.snapcraft.io/"
STORE_API_URL = "https://api.snapcraft.io/"
STORE_UPLOAD_URL = "https://upload.apps.ubuntu.com/"

# Messages and warnings.
MISSING_AGREEMENT = "Developer has not signed agreement."
MISSING_NAMESPACE = "Developer profile is missing short namespace."
AGREEMENT_ERROR = (
    "You must agree to the developer terms and conditions to upload snaps."
)
NAMESPACE_ERROR = (
    "You need to set a username. It will appear in the developer field "
    "alongside the other details for your snap. Please visit {} and login "
    "again."
)
AGREEMENT_INPUT_MSG = "Do you agree to the developer terms and conditions. ({})?"
AGREEMENT_SIGN_ERROR = (
    "Unexpected error encountered during signing the developer terms and "
    "conditions. Please visit {} and agree to the terms and conditions before "
    "continuing."
)
TWO_FACTOR_WARNING = (
    "We strongly recommend enabling multi-factor authentication: "
    "https://help.ubuntu.com/community/SSO/FAQs/2FA"
)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import configparser
import enum
import logging
import os
from typing import Optional

from xdg import BaseDirectory

from snapcraft.internal.errors import SnapcraftInvalidCLIConfigError


logger = logging.getLogger(__name__)


@enum.unique
class OutdatedStepAction(enum.Enum):
    # Would like to use enum.auto(), but it's only available in >= 3.6
    ERROR = 1
    CLEAN = 2


class CLIConfig:
    """Hold general options affecting the CLI.

    Setting data takes the convention of set_<section>_<option> and
    getting data takes a similar form get_<section>_<option>.
    """

    def __init__(self, *, read_only: bool = False) -> None:
        """Initialize an instance of CLIConfig.

        :param bool read_only: if set, attempts to set data will fail.
        """
        self.parser = configparser.ConfigParser()
        self.config_path = os.path.join(
            BaseDirectory.save_config_path("snapcraft"), "cli.cfg"
        )
        self._read_only = read_only

    def __enter__(self):
        self.load()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self._read_only:
            self.save()

    def load(self) -> None:
        """Load config parser data.

        :raises SnapcraftInvalidCLIConfigError:
            if the configuration data cannot be loaded from disk.
        """
        config: Optional[str] = None

        try:
            with open(self.config_path) as f:
                config = f.read()
        except FileNotFoundError:
            config = None

        if config:
            try:
                self.parser.read_string(config)
            except configparser.Error as parse_error:
                raise SnapcraftInvalidCLIConfigError(
                    config_file=self.config_path, error=str(parse_error)
                ) from parse_error

    def save(self) -> None:
        """Save config parser data.

        :raises RuntimeError: if trying to save when operating with read_only.
        """
        if self._read_only:
            raise RuntimeError("CLIConfig was instantiated with read_only.")
        # Make sure the directory exists before saving
        os.makedirs(os.path.dirname(self.config_path), exist_ok=True)

        with open(self.config_path, "w") as f:
            self.parser.write(f)

    def _get_option(self, section_name: str, option_name: str) -> Optional[str]:
        try:
            return self.parser.get(section_name, option_name)
        except (configparser.NoSectionError, configparser.NoOptionError, KeyError):
            return None

    def _set_option(self, section_name, option_name: str, value: str) -> None:
        if self._read_only:
            raise RuntimeError("CLIConfig was instantiated with read_only.")

        section_name = section_name
        if not self.parser.has_section(section_name):
            self.parser.add_section(section_name)
        self.parser.set(section_name, option_name, value)

    def set_sentry_send_always(self, value: bool) -> None:
        """Setter to define if errors should automatically be sent to sentry.

        :param bool value: if true, errors should automatically be sent.
        """
        if value:
            string_value = "true"
        else:
            string_value = "false"
        self._set_option("Sentry", "always_send", string_value)

    def get_sentry_send_always(self) -> bool:
        """Getter to define if errors should automatically be sent to sentry.

        :returns: True if errors should automatically be sent.
        :rtype: bool.
        """
        string_value = self._get_option("Sentry", "always_send")
        # Anything but "true" for string_value is considered False.
        return string_value == "true"

    def set_outdated_step_action(self, action: OutdatedStepAction) -> None:
        """Setter to define action to take if outdated step is encountered.

        :param OutdatedStepAction value: The action to take
        """
        self._set_option("Lifecycle", "outdated_step_action", action.name.lower())

    def get_outdated_step_action(self) -> OutdatedStepAction:
        """Getter to define action to take if outdated step is encountered.

        :returns: The action to take
        :rtype: OutdatedStepAction.
        """
        action = self._get_option("Lifecycle", "outdated_step_action")
        if action:
            return OutdatedStepAction[action.upper()]
        else:
            # Clean by default
            return OutdatedStepAction.CLEAN



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os

from tests import unit


# LP: #1733584
class PythonBaseTestCase(unit.TestCase):  # type: ignore
    def _create_python_binary(self, base_dir):
        python_command_path = os.path.join(base_dir, "usr", "bin", "pythontest")
        os.makedirs(os.path.dirname(python_command_path), exist_ok=True)
        open(python_command_path, "w").close()
        return python_command_path



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019-2021 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import pathlib

from testtools.matchers import Contains, Equals, FileExists, Not

from snapcraft import yaml_utils
from snapcraft.internal.meta import application, desktop, errors
from tests import unit


class AppCommandTest(unit.TestCase):
    def setUp(self):
        super().setUp()
        for exe in (
            "test-command",
            "test-stop-command",
            "test-command-chain",
            "prepend-command-chain",
        ):
            open(exe, "w").close()
            os.chmod(exe, 0o755)

    def test_app_no_change(self):
        app = application.Application.from_dict(
            app_name="foo",
            app_dict={
                "command": "test-command",
                "stop-command": "test-stop-command",
                "daemon": "simple",
                "install-mode": "disable",
                "command-chain": ["test-command-chain"],
            },
        )

        app.prime_commands(base="core18", prime_dir=self.path)

        self.expectThat(
            app.to_dict(),
            Equals(
                {
                    "command": "test-command",
                    "stop-command": "test-stop-command",
                    "daemon": "simple",
                    "install-mode": "disable",
                    "command-chain": ["test-command-chain"],
                }
            ),
        )

        app.write_command_wrappers(prime_dir=self.path)
        self.expectThat("command-foo.wrapper", Not(FileExists()))
        self.expectThat("stop-command-foo.wrapper", Not(FileExists()))

    def test_app_with_wrapper(self):
        app = application.Application.from_dict(
            app_name="foo",
            app_dict={
                "command": "/test-command",
                "stop-command": "/test-stop-command",
                "daemon": "simple",
            },
        )
        app.prime_commands(base="core18", prime_dir=self.path)
        self.assertThat(
            app.to_dict(),
            Equals(
                {
                    "command": "command-foo.wrapper",
                    "stop-command": "stop-command-foo.wrapper",
                    "daemon": "simple",
                }
            ),
        )

        app.write_command_wrappers(prime_dir=self.path)
        self.expectThat("command-foo.wrapper", FileExists())
        self.expectThat("stop-command-foo.wrapper", FileExists())

    def test_massaged_core(self):
        app = application.Application.from_dict(
            app_name="foo", app_dict={"command": "$SNAP/test-command"}
        )
        app.prime_commands(base="core", prime_dir=self.path)

        self.assertThat(app.to_dict(), Equals({"command": "test-command"}))

    def test_massaged_core18(self):
        app = application.Application.from_dict(
            app_name="foo", app_dict={"command": "$SNAP/test-command"}
        )
        app.prime_commands(base="core18", prime_dir=self.path)

        self.assertThat(app.to_dict(), Equals({"command": "test-command"}))

    def test_not_massaged_core20(self):
        app = application.Application.from_dict(
            app_name="foo", app_dict={"command": "$SNAP/test-command"}
        )

        self.assertRaises(
            errors.InvalidAppCommandNotFound,
            app.prime_commands,
            base="core20",
            prime_dir=self.path,
        )

    def test_socket_mode_change_to_octal(self):
        app = application.Application.from_dict(
            app_name="foo",
            app_dict={
                "command": "test-command",
                "daemon": "simple",
                "sockets": {
                    "sock1": {"listen-stream": 8080},
                    "sock2": {
                        "listen-stream": "$SNAP_COMMON/sock2",
                        "socket-mode": 1000,
                    },
                },
            },
        )

        self.expectThat(
            type(app.to_dict()["sockets"]["sock2"]["socket-mode"]),
            Equals(yaml_utils.OctInt),
        )

    def test_no_command_chain(self):
        app = application.Application.from_dict(
            app_name="foo", app_dict={"command": "test-command"}
        )

        app.prime_commands(base="core18", prime_dir=self.path)

        self.assertThat(app.to_dict(), Equals({"command": "test-command"}))


class TestWrapperUse:
    scenarios = (
        (
            "wrapper allowed for plain command on core18",
            dict(extra_app_properties={}, base="core18", expect_wrappers=True),
        ),
        (
            "wrapper allowed for plain command on core",
            dict(extra_app_properties={}, base="core", expect_wrappers=True),
        ),
        (
            "wrapper not allowed for not core or core18 base",
            dict(extra_app_properties={}, base="core20", expect_wrappers=False),
        ),
        (
            "wrapper not allowed with command-chain",
            dict(
                extra_app_properties={"command-chain": ["command-chain"]},
                base="core18",
                expect_wrappers=False,
            ),
        ),
        (
            "wrapper not allowed with none adapter",
            dict(
                extra_app_properties={"adapter": "none"},
                base="core18",
                expect_wrappers=False,
            ),
        ),
    )

    def test_wrapper(self, tmp_work_path, extra_app_properties, base, expect_wrappers):
        app_properties = dict(command="foo")
        app_properties.update(extra_app_properties)

        for exe in ["foo"] + app_properties.get("command-chain", list()):
            exe_path = pathlib.Path(exe)
            exe_path.touch()
            exe_path.chmod(0o755)

        app = application.Application.from_dict(app_name="foo", app_dict=app_properties)

        assert app.can_use_wrapper(base=base) == expect_wrappers


class InvalidCommandChainTest(unit.TestCase):
    def test_command_chain_path_not_found(self):
        app = application.Application.from_dict(
            app_name="foo", app_dict={"command-chain": "file-not-found"}
        )

        self.assertRaises(
            errors.InvalidCommandChainError,
            app.validate_command_chain_executables,
            prime_dir=self.path,
        )

    def test_command_chain_path_not_executable(self):
        open("file-not-executable", "w").close()

        app = application.Application.from_dict(
            app_name="foo", app_dict={"command-chain": "file-not-executable"}
        )

        self.assertRaises(
            errors.InvalidCommandChainError,
            app.validate_command_chain_executables,
            prime_dir=self.path,
        )


class DesktopFileTest(unit.TestCase):
    def test_desktop_file(self):
        desktop_file_path = "foo.desktop"
        with open(desktop_file_path, "w") as desktop_file:
            print("[Desktop Entry]", file=desktop_file)
            print("Exec=in-snap-exe", file=desktop_file)
        open("command-chain", "w").close()
        os.chmod("command-chain", 0o755)

        app = application.Application.from_dict(
            app_name="foo", app_dict=dict(command="/foo", desktop=desktop_file_path)
        )

        desktop_file = desktop.DesktopFile(
            snap_name="foo",
            app_name=app.app_name,
            filename=app.desktop,
            prime_dir=self.path,
        )

        desktop_file.write(gui_dir="gui")

        expected_desktop_file_path = os.path.join("gui", "foo.desktop")

        self.expectThat(app.to_dict(), Not(Contains("desktop")))
        self.expectThat(expected_desktop_file_path, FileExists())


class AppPassthroughTests(unit.TestCase):
    def test_no_passthrough(self):
        app = application.Application(
            app_name="foo",
            adapter=application.ApplicationAdapter.NONE,
            command_chain=["test-command-chain"],
            passthrough=None,
        )

        app_dict = app.to_dict()

        self.assertThat(app_dict, Equals({"command-chain": ["test-command-chain"]}))

    def test_passthrough_to_dict(self):
        app = application.Application(
            app_name="foo",
            adapter=application.ApplicationAdapter.NONE,
            command_chain=["test-command-chain"],
            passthrough={"test-property": "test-value"},
        )

        app_dict = app.to_dict()

        self.assertThat(
            app_dict,
            Equals(
                {"command-chain": ["test-command-chain"], "test-property": "test-value"}
            ),
        )

    def test_passthrough_to_dict_from_dict(self):
        app = application.Application.from_dict(
            app_name="foo",
            app_dict={
                "adapter": "none",
                "command-chain": ["test-command-chain"],
                "passthrough": {"test-property": "test-value"},
            },
        )

        app_dict = app.to_dict()

        self.assertThat(
            app_dict,
            Equals(
                {"command-chain": ["test-command-chain"], "test-property": "test-value"}
            ),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
from unittest import mock

from testtools.matchers import Equals

from snapcraft.internal import xattrs
from snapcraft.internal.errors import XAttributeTooLongError
from tests import unit


class TestXattrs(unit.TestCase):
    def setUp(self):
        super().setUp()
        self.test_file = os.path.join(self.path, "test-file")
        open(self.test_file, "w").close()

    def test_read_origin_stage_package(self):
        if sys.platform == "linux":
            result = xattrs.read_origin_stage_package(self.test_file)
            self.assertThat(result, Equals(None))
        else:
            self.assertRaises(
                RuntimeError, xattrs.read_origin_stage_package, self.test_file
            )

    def test_write_origin_stage_package(self):
        package = "foo-1.0"
        if sys.platform == "linux":
            result = xattrs.read_origin_stage_package(self.test_file)
            self.assertThat(result, Equals(None))

            xattrs.write_origin_stage_package(self.test_file, package)
            result = xattrs.read_origin_stage_package(self.test_file)
            self.assertThat(result, Equals(package))
        else:
            self.assertRaises(
                RuntimeError, xattrs.write_origin_stage_package, self.test_file, package
            )

    def test_write_origin_stage_package_long(self):
        package = "a" * 100000
        if sys.platform == "linux":
            result = xattrs.read_origin_stage_package(self.test_file)
            self.assertThat(result, Equals(None))

            self.assertRaises(
                XAttributeTooLongError,
                xattrs.write_origin_stage_package,
                self.test_file,
                package,
            )

            result = xattrs.read_origin_stage_package(self.test_file)
            self.assertThat(result, Equals(None))
        else:
            self.assertRaises(
                RuntimeError, xattrs.write_origin_stage_package, self.test_file, package
            )

    def test_symlink(self):
        test_symlink = self.test_file + "-symlink"
        os.symlink(self.test_file, test_symlink)

        if sys.platform != "linux":
            return

        result = xattrs._read_snapcraft_xattr(test_symlink, "attr")
        self.assertThat(result, Equals(None))

        xattrs._write_snapcraft_xattr(test_symlink, "attr", "value")
        result = xattrs._read_snapcraft_xattr(test_symlink, "attr")
        self.assertThat(result, Equals(None))

    @mock.patch("sys.platform", return_value="win32")
    def test_read_non_linux(self, mock_platform):
        self.assertRaises(
            RuntimeError, xattrs._read_snapcraft_xattr, self.test_file, "attr"
        )

    @mock.patch("sys.platform", return_value="win32")
    def test_write_non_linux(self, mock_platform):
        self.assertRaises(
            RuntimeError, xattrs._write_snapcraft_xattr, self.test_file, "attr", "value"
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018,2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import copy
import pathlib
import sys
from textwrap import dedent

import fixtures
from testtools.matchers import Equals, IsInstance

from snapcraft.internal import errors
from snapcraft.plugins._plugin_finder import _PLUGINS
from snapcraft.plugins.v1 import PluginV1
from snapcraft.plugins.v2 import PluginV2
from tests import unit


class NonLocalTest(unit.TestCase):
    def test_unknown_plugin_must_raise_exception(self):
        path = copy.copy(sys.path)

        raised = self.assertRaises(
            errors.PluginError, self.load_part, "fake-part", "not-found"
        )

        self.assertThat(raised.message, Equals("unknown plugin: 'not-found'"))

        # Make sure that nothing was added to sys.path.
        self.assertThat(path, Equals(sys.path))

    def test_local_plugin(self):
        plugin_path = pathlib.Path("snap/plugins/x_local.py")
        plugin_path.parent.mkdir(parents=True)
        with open(plugin_path, "w") as plugin:
            print(
                dedent(
                    """\
                import snapcraft.plugins.v1
                class Local(snapcraft.plugins.v1.PluginV1):
                    pass
                """
                ),
                file=plugin,
            )

        self.load_part("test-part", plugin_name="x-local")


class InTreePluginsTest(unit.TestCase):
    def test_all_known_v1(self):
        # We don't want validation to take place here.
        self.useFixture(fixtures.MockPatch("jsonschema.validate"))
        for plugin_name in _PLUGINS["v1"]:
            plugin_handler = self.load_part(
                "test-part", plugin_name=plugin_name, base="core18"
            )
            self.expectThat(plugin_handler.plugin, IsInstance(PluginV1))

    def test_all_v2(self):
        self.useFixture(fixtures.MockPatch("jsonschema.validate"))
        for plugin_name in _PLUGINS["v2"]:
            plugin_handler = self.load_part(
                "test-part", plugin_name=plugin_name, base="core20"
            )
            self.expectThat(plugin_handler.plugin, IsInstance(PluginV2))

    def test_fail_on_schema(self):
        # conda requires some part_properties to be set.
        self.assertRaises(
            errors.PluginError,
            self.load_part,
            "test-part",
            plugin_name="conda",
            base="core18",
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import textwrap

from snapcraft.internal import lifecycle, states, steps
from snapcraft.internal.lifecycle._status_cache import StatusCache

from . import LifecycleTestBase


class StatusCacheTestCase(LifecycleTestBase):
    def setUp(self):
        super().setUp()

        self.project_config = self.make_snapcraft_project(
            textwrap.dedent(
                """\
                parts:
                  main:
                    source: .
                    plugin: dump
                  dependent:
                    plugin: nil
                    after: [main]
                """
            )
        )

        self.cache = StatusCache(self.project_config)

    def test_has_step_run(self):
        # No steps should have run, yet
        main_part = self.project_config.parts.get_part("main")
        self.assertFalse(self.cache.has_step_run(main_part, steps.PULL))

        # Now run the pull step
        lifecycle.execute(steps.PULL, self.project_config, part_names=["main"])

        # Should still have cached that no steps have run
        self.assertFalse(self.cache.has_step_run(main_part, steps.PULL))

        # Now clear that step from the cache, and it should be up-to-date
        self.cache.clear_step(main_part, steps.PULL)
        self.assertTrue(self.cache.has_step_run(main_part, steps.PULL))

    def test_add_step_run(self):
        # No steps should have run, yet
        main_part = self.project_config.parts.get_part("main")
        self.assertFalse(self.cache.has_step_run(main_part, steps.PULL))

        # Tell the cache that the pull step has run though
        self.cache.add_step_run(main_part, steps.PULL)

        # Now it should think that it has actually run
        self.assertTrue(self.cache.has_step_run(main_part, steps.PULL))

        # Now clear that step from the cache, and it should no longer have ran
        self.cache.clear_step(main_part, steps.PULL)
        self.assertFalse(self.cache.has_step_run(main_part, steps.PULL))

    def test_get_dirty_report(self):
        # No dirty reports should be available, yet
        dependent_part = self.project_config.parts.get_part("dependent")
        self.assertFalse(self.cache.get_dirty_report(dependent_part, steps.PULL))

        # Now run the pull step
        lifecycle.execute(steps.PULL, self.project_config)

        # Re-stage main, which will make dependent dirty
        lifecycle.execute(steps.PULL, self.project_config, part_names=["main"])

        # Should still have cached that it's not dirty, though
        self.assertFalse(self.cache.get_dirty_report(dependent_part, steps.PULL))

        # Now clear that step from the cache, and it should be up-to-date
        self.cache.clear_step(dependent_part, steps.PULL)
        self.assertTrue(self.cache.get_dirty_report(dependent_part, steps.PULL))

    def test_get_outdated_report(self):
        # No outdated reports should be available, yet
        main_part = self.project_config.parts.get_part("main")
        self.assertFalse(self.cache.get_outdated_report(main_part, steps.PULL))

        # Now run the pull step for main
        lifecycle.execute(steps.PULL, self.project_config, part_names=["main"])

        # Change the source on disk, which will make the pull step of main
        # outdated (to ensure this is the case, manually set the timestamp)
        open("new-file", "w").close()
        pull_state_file = states.get_step_state_file(
            main_part.plugin.statedir, steps.PULL
        )
        access_time = os.stat(pull_state_file).st_atime
        modified_time = os.stat(pull_state_file).st_atime
        os.utime("new-file", (access_time, modified_time + 1))

        # Should still have cached that it's not outdated, though
        self.assertFalse(self.cache.get_outdated_report(main_part, steps.PULL))

        # Now clear that step from the cache, and it should be up-to-date
        self.cache.clear_step(main_part, steps.PULL)
        self.assertTrue(self.cache.get_outdated_report(main_part, steps.PULL))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import stat

import yaml as yaml_utils
from testtools.matchers import Equals, FileExists

from tests import integration


# FIXME: Ideally this would be a snaps test so we can run the aliases, but that
# requires snapd v2.20 to be available everywhere (it's not). We'll just make
# sure it builds instead.
class AliasTestCase(integration.TestCase):
    def test_alias(self):
        self.run_snapcraft("prime", "alias")
        file_path = os.path.join(self.prime_dir, "bin", "hello.sh")
        self.assertThat(file_path, FileExists())
        self.assertTrue(
            os.stat(file_path).st_mode & stat.S_IEXEC,
            "Expected hello.sh to be executable",
        )

        snap_yaml = os.path.join(self.prime_dir, "meta", "snap.yaml")
        self.assertThat(snap_yaml, FileExists())

        data = {}
        with open(snap_yaml) as fp:
            data = yaml_utils.load(fp)

        expected_aliases = ["hi.sh", "howdy.sh"]
        self.assertThat(
            set(data["apps"]["hello"]["aliases"]), Equals(set(expected_aliases))
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import fileinput
import glob
import itertools
import logging
import os
import pathlib
import re
import shutil
import stat
from typing import List, Optional, Set

from snapcraft import file_utils
from snapcraft.internal import mangling, xattrs

from . import errors

logger = logging.getLogger(__name__)


class BaseRepo:
    """Base implementation for a platform specific repo handler.

    Generally any new repo handling system would inherit from this and
    implement:

    - get
    - unpack
    - get_package_libraries
    - get_packages_for_source_type
    - refresh_build_packages
    - install_build_packages
    - is_package_installed

    At the end of the `unpack` method `normalize` needs to be called to
    adapt the artifacts downloaded to be generic enough for building a snap."""

    @classmethod
    def get_package_libraries(cls, package_name: str) -> Set[str]:
        """Return a list of libraries in package_name.

        Given the contents of package_name, return the subset of what are
        considered libraries from those contents, be it static or shared.

        :param str package: package name to get library contents from.
        :returns: a list of libraries that package_name provides. This includes
                  directories.
        :rtype: set.
        """
        raise errors.NoNativeBackendError()

    @classmethod
    def get_package_for_file(cls, file_path: str) -> str:
        """Return the package name that provides file_path.

        :param str file_path: the absolute path to the file to search for.
        :returns: package name that provides file_path.
        :rtype: str
        :raises snapcraft.repo.errors.FileProviderNotFound:
            if file_path is not provided by any package.
        """
        raise errors.NoNativeBackendError()

    @classmethod
    def get_packages_for_source_type(cls, source_type: str) -> Set[str]:
        """Return a list of packages required to to work with source_type.

        source_type can be any of the following:

        - bzr
        - deb
        - rpm
        - git
        - hg
        - mercurial
        - subversion
        - svn
        - tar
        - zip

        :param str source_type: a VCS source type to handle.
        :returns: a set of packages that need to be installed on the host.
        :rtype: set of strings.
        """
        raise errors.NoNativeBackendError()

    @classmethod
    def refresh_build_packages(cls) -> None:
        """Refresh the build packages cache.

        If refreshing is not possible
        snapcraft.repo.errors.CacheUpdateFailedError should be raised

        :raises snapcraft.repo.errors.NoNativeBackendError:
            if the method is not implemented in the subclass.
        """
        raise errors.NoNativeBackendError()

    @classmethod
    def install_build_packages(cls, package_names: List[str]) -> List[str]:
        """Install packages on the host required to build.

        This method needs to be implemented by using the appropriate method
        to install packages on the system. If possible they should be marked
        as automatically installed to allow for easy removal.
        The method should return a list of the actually installed packages
        in the form "package=version".

        If one of the packages cannot be found
        snapcraft.repo.errors.BuildPackageNotFoundError should be raised.
        If dependencies for a package cannot be resolved
        snapcraft.repo.errors.PackageBrokenError should be raised.
        If installing a package on the host failed
        snapcraft.repo.errors.BuildPackagesNotInstalledError should be raised.

        :param package_names: a list of package names to install.
        :type package_names: a list of strings.
        :return: a list with the packages installed and their versions.
        :rtype: list of strings.
        :raises snapcraft.repo.errors.NoNativeBackendError:
            if the method is not implemented in the subclass.
        """
        raise errors.NoNativeBackendError()

    @classmethod
    def build_package_is_valid(cls, package_name: str) -> bool:
        """Check that a given package is valid on the host.

        :param package_name: a package name to check.
        :type package_name: str
        """
        raise errors.NoNativeBackendError()

    @classmethod
    def is_package_installed(cls, package_name: str) -> bool:
        """Return a bool indicating if package_name is installed.

        :param str package_name: the package name to query.
        :returns: True if package_name is installed if not False.
        :rtype: boolean
        """
        raise errors.NoNativeBackendError()

    @classmethod
    def get_installed_packages(cls) -> List[str]:
        """Return a list of the installed packages and their versions

        :rtype: list of strings with the form package=version.
        """
        raise errors.NoNativeBackendError()

    @classmethod
    def fetch_stage_packages(
        cls,
        *,
        package_names: List[str],
        base: str,
        stage_packages_path: pathlib.Path,
        target_arch: str,
    ) -> List[str]:
        """Fetch stage packages to stage_packages_path."""
        raise errors.NoNativeBackendError()

    @classmethod
    def unpack_stage_packages(
        cls, *, stage_packages_path: pathlib.Path, install_path: pathlib.Path
    ) -> None:
        """Unpack stage packages to install_path."""
        raise errors.NoNativeBackendError()

    @classmethod
    def install_gpg_key(cls, *, key_id: str, key: str) -> bool:
        """Install trusted GPG key."""
        raise errors.NoNativeBackendError()

    @classmethod
    def normalize(cls, unpackdir: str) -> None:
        """Normalize artifacts in unpackdir.

        Repo specific packages are generally created to live in a specific
        distro. What normalize does is scan through the unpacked artifacts
        and slightly modifies them to work better with snapcraft projects
        when building and to also work within a snap's environment.

        :param str unpackdir: directory where files where unpacked.
        """
        cls._remove_useless_files(unpackdir)
        cls._fix_artifacts(unpackdir)
        cls._fix_xml_tools(unpackdir)
        cls._fix_shebangs(unpackdir)

    @classmethod
    def _mark_origin_stage_package(
        cls, sources_dir: str, stage_package: str
    ) -> Set[str]:
        """Mark all files in sources_dir as coming from stage_package."""
        file_list = set()
        for (root, dirs, files) in os.walk(sources_dir):
            for file_name in files:
                file_path = os.path.join(root, file_name)

                # Mark source.
                xattrs.write_origin_stage_package(file_path, stage_package)

                file_path = os.path.relpath(root, sources_dir)
                file_list.add(file_path)

        return file_list

    @classmethod
    def _remove_useless_files(cls, unpackdir: str) -> None:
        """Remove files that aren't useful or will clash with other parts."""
        sitecustomize_files = glob.glob(
            os.path.join(unpackdir, "usr", "lib", "python*", "sitecustomize.py")
        )
        for sitecustomize_file in sitecustomize_files:
            os.remove(sitecustomize_file)

    @classmethod
    def _fix_artifacts(cls, unpackdir: str) -> None:
        """Perform various modifications to unpacked artifacts.

        Sometimes distro packages will contain absolute symlinks (e.g. if the
        relative path would go all the way to root, they just do absolute). We
        can't have that, so instead clean those absolute symlinks.

        Some unpacked items will also contain suid binaries which we do not
        want in the resulting snap.
        """
        for root, dirs, files in os.walk(unpackdir):
            # Symlinks to directories will be in dirs, while symlinks to
            # non-directories will be in files.
            for entry in itertools.chain(files, dirs):
                path = os.path.join(root, entry)
                if os.path.islink(path):
                    cls._fix_symlink(path, unpackdir)
                elif os.path.exists(path):
                    _fix_filemode(path)

                if (
                    path.endswith(".pc")
                    and os.path.isfile(path)
                    and not os.path.islink(path)
                ):
                    fix_pkg_config(unpackdir, path)

    @classmethod
    def _fix_xml_tools(cls, unpackdir: str) -> None:
        xml2_config_path = os.path.join(unpackdir, "usr", "bin", "xml2-config")
        with contextlib.suppress(FileNotFoundError):
            file_utils.search_and_replace_contents(
                xml2_config_path,
                re.compile(r"prefix=/usr"),
                "prefix={}/usr".format(unpackdir),
            )

        xslt_config_path = os.path.join(unpackdir, "usr", "bin", "xslt-config")
        with contextlib.suppress(FileNotFoundError):
            file_utils.search_and_replace_contents(
                xslt_config_path,
                re.compile(r"prefix=/usr"),
                "prefix={}/usr".format(unpackdir),
            )

    @classmethod
    def _fix_symlink(cls, symlink_path: str, unpack_dir: str) -> None:
        """Verify, fix, and/or log symlinks that would break in snap."""
        symlink_target = os.readlink(symlink_path)
        resolved_path = os.path.realpath(symlink_path)

        # If symlink doesn't point to absolute target, nothing to do.
        if not os.path.isabs(symlink_target):
            return

        # Ignore anything linking to libc6.
        if resolved_path in cls.get_package_libraries("libc6"):
            logger.debug("Not fixing symlink %r: it's pointing to libc", symlink_target)
            return

        # Use relative symlink if target is found in unpack directory.
        target_in_unpack_dir = os.path.join(unpack_dir, symlink_target[1:])
        relative_target = os.path.relpath(
            target_in_unpack_dir, os.path.dirname(symlink_path)
        )
        # Copy the file from host if we need it, but warn about it.
        if not os.path.exists(target_in_unpack_dir) and os.path.isfile(resolved_path):
            logger.warning(
                "Copying needed symlink target %r from host to satisfy %r.",
                resolved_path,
                symlink_path,
            )
            os.makedirs(os.path.dirname(target_in_unpack_dir), exist_ok=True)
            shutil.copyfile(resolved_path, target_in_unpack_dir)

        # Relink to relative path.
        if os.path.exists(target_in_unpack_dir):
            os.unlink(symlink_path)
            os.symlink(relative_target, symlink_path)
            return

        # Cannot find target, issue a warning.
        relative_link = os.path.relpath(symlink_path, unpack_dir)
        logger.warning("%r will be a dangling symlink", relative_link)

    @classmethod
    def _fix_shebangs(cls, unpackdir: str) -> None:
        """Change hard-coded shebangs in unpacked files to use env."""
        mangling.rewrite_python_shebangs(unpackdir)


class DummyRepo(BaseRepo):
    @classmethod
    def get_packages_for_source_type(cls, source_type: str) -> Set[str]:
        return set()


def fix_pkg_config(
    root: str, pkg_config_file: str, prefix_trim: Optional[str] = None
) -> None:
    """Opens a pkg_config_file and prefixes the prefix with root."""
    pattern_trim = None
    if prefix_trim:
        pattern_trim = re.compile("^prefix={}(?P<prefix>.*)".format(prefix_trim))
    pattern = re.compile("^prefix=(?P<prefix>.*)")

    with fileinput.input(pkg_config_file, inplace=True) as input_file:
        for line in input_file:
            match = pattern.search(line)
            if prefix_trim is not None and pattern_trim is not None:
                match_trim = pattern_trim.search(line)
            if prefix_trim is not None and match_trim is not None:
                print("prefix={}{}".format(root, match_trim.group("prefix")))
            elif match:
                print("prefix={}{}".format(root, match.group("prefix")))
            else:
                print(line, end="")


def _fix_filemode(path: str) -> None:
    mode = stat.S_IMODE(os.stat(path, follow_symlinks=False).st_mode)
    if mode & 0o4000 or mode & 0o2000:
        logger.warning("Removing suid/guid from {}".format(path))
        os.chmod(path, mode & 0o1777)


def get_pkg_name_parts(pkg_name):
    """Break package name into base parts"""

    name = pkg_name
    version = None
    with contextlib.suppress(ValueError):
        name, version = pkg_name.split("=")

    return name, version



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from unittest import mock

from testtools.matchers import FileExists

from snapcraft.internal.sources import Script
from tests import unit


class TestScript(unit.TestCase):
    def setUp(self):
        super().setUp()
        self.source = Script("source", "destination")

        os.mkdir("destination")
        self.source.file = os.path.join("destination", "file")
        open(self.source.file, "w").close()

    @mock.patch("snapcraft.internal.sources._script.FileBase.download")
    def test_download_makes_executable(self, mock_download):
        self.source.file = os.path.join("destination", "file")
        self.source.download()
        self.assertThat(self.source.file, FileExists())
        self.assertThat(self.source.file, unit.IsExecutable())



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from testtools.matchers import Equals

from snapcraft.internal.project_loader._extensions.gnome_3_38 import ExtensionImpl
from tests.unit.commands import CommandBaseTestCase

from .. import ProjectLoaderBaseTest


class ExtensionTest(ProjectLoaderBaseTest, CommandBaseTestCase):
    def test_extension(self):
        gnome_extension = ExtensionImpl(
            extension_name="gnome-3.38", yaml_data=dict(base="core20")
        )

        self.expectThat(
            gnome_extension.root_snippet,
            Equals(
                {
                    "assumes": ["snapd2.43"],
                    "plugs": {
                        "desktop": {"mount-host-font-cache": False},
                        "gtk-3-themes": {
                            "interface": "content",
                            "target": "$SNAP/data-dir/themes",
                            "default-provider": "gtk-common-themes",
                        },
                        "icon-themes": {
                            "interface": "content",
                            "target": "$SNAP/data-dir/icons",
                            "default-provider": "gtk-common-themes",
                        },
                        "sound-themes": {
                            "interface": "content",
                            "target": "$SNAP/data-dir/sounds",
                            "default-provider": "gtk-common-themes",
                        },
                        "gnome-3-38-2004": {
                            "interface": "content",
                            "target": "$SNAP/gnome-platform",
                            "default-provider": "gnome-3-38-2004",
                        },
                    },
                    "environment": {
                        "SNAP_DESKTOP_RUNTIME": "$SNAP/gnome-platform",
                        "GTK_USE_PORTAL": "1",
                    },
                    "hooks": {
                        "configure": {
                            "plugs": ["desktop"],
                            "command-chain": [
                                "snap/command-chain/hooks-configure-desktop"
                            ],
                        }
                    },
                    "layout": {
                        "/usr/lib/$SNAPCRAFT_ARCH_TRIPLET/webkit2gtk-4.0": {
                            "bind": "$SNAP/gnome-platform/usr/lib/$SNAPCRAFT_ARCH_TRIPLET/webkit2gtk-4.0"
                        },
                        "/usr/share/xml/iso-codes": {
                            "bind": "$SNAP/gnome-platform/usr/share/xml/iso-codes"
                        },
                    },
                }
            ),
        )
        self.expectThat(
            gnome_extension.app_snippet,
            Equals(
                {
                    "command-chain": ["snap/command-chain/desktop-launch"],
                    "plugs": [
                        "desktop",
                        "desktop-legacy",
                        "gsettings",
                        "opengl",
                        "wayland",
                        "x11",
                    ],
                }
            ),
        )
        self.expectThat(
            gnome_extension.part_snippet,
            Equals(
                {
                    "build-environment": [
                        {"PATH": "/snap/gnome-3-38-2004-sdk/current/usr/bin:$PATH"},
                        {
                            "XDG_DATA_DIRS": "$SNAPCRAFT_STAGE/usr/share:/snap/gnome-3-38-2004-sdk/current/usr/share:/usr/share:$XDG_DATA_DIRS"
                        },
                        {
                            "LD_LIBRARY_PATH": ":".join(
                                [
                                    "/snap/gnome-3-38-2004-sdk/current/lib/$SNAPCRAFT_ARCH_TRIPLET",
                                    "/snap/gnome-3-38-2004-sdk/current/usr/lib/$SNAPCRAFT_ARCH_TRIPLET",
                                    "/snap/gnome-3-38-2004-sdk/current/usr/lib",
                                    "/snap/gnome-3-38-2004-sdk/current/usr/lib/vala-current",
                                    "/snap/gnome-3-38-2004-sdk/current/usr/lib/$SNAPCRAFT_ARCH_TRIPLET/pulseaudio",
                                ]
                            )
                            + "${LD_LIBRARY_PATH:+:$LD_LIBRARY_PATH}"
                        },
                        {
                            "PKG_CONFIG_PATH": "/snap/gnome-3-38-2004-sdk/current/usr/lib/$SNAPCRAFT_ARCH_TRIPLET/pkgconfig:/snap/gnome-3-38-2004-sdk/current/usr/lib/pkgconfig:/snap/gnome-3-38-2004-sdk/current/usr/share/pkgconfig:$PKG_CONFIG_PATH"
                        },
                        {
                            "GETTEXTDATADIRS": "/snap/gnome-3-38-2004-sdk/current/usr/share/gettext-current:$GETTEXTDATADIRS"
                        },
                        {
                            "GDK_PIXBUF_MODULE_FILE": "/snap/gnome-3-38-2004-sdk/current/usr/lib/$SNAPCRAFT_ARCH_TRIPLET/gdk-pixbuf-current/loaders.cache"
                        },
                        {
                            "ACLOCAL_PATH": "/snap/gnome-3-38-2004-sdk/current/usr/share/aclocal${ACLOCAL_PATH:+:$ACLOCAL_PATH}"
                        },
                        {
                            "PYTHONPATH": ":".join(
                                [
                                    "/snap/gnome-3-38-2004-sdk/current/usr/lib/python3.8",
                                    "/snap/gnome-3-38-2004-sdk/current/usr/lib/python3/dist-packages",
                                ]
                            )
                            + "${PYTHONPATH:+:$PYTHONPATH}"
                        },
                    ]
                }
            ),
        )
        self.expectThat(
            gnome_extension.parts,
            Equals(
                {
                    "gnome-3-38-extension": {
                        "source": "$SNAPCRAFT_EXTENSIONS_DIR/desktop",
                        "source-subdir": "gnome",
                        "plugin": "make",
                        "make-parameters": [
                            "WITH_PYTHON=3.8",
                            "PLATFORM_PLUG=gnome-3-38-2004",
                        ],
                        "build-snaps": ["gnome-3-38-2004-sdk/latest/stable"],
                        "build-packages": ["gcc"],
                    }
                }
            ),
        )

    def test_supported_bases(self):
        self.assertThat(ExtensionImpl.get_supported_bases(), Equals(("core20",)))

    def test_supported_confinement(self):
        self.assertThat(
            ExtensionImpl.get_supported_confinement(), Equals(("strict", "devmode"))
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import operator
import os
from io import StringIO
from typing import List, Optional

import lxml.etree
from xdg.DesktopEntry import DesktopEntry

from snapcraft.extractors import _errors

from ._metadata import ExtractedMetadata

_xslt = None
_XSLT = """\
<?xml version="1.0"?>
<xsl:stylesheet version="1.0"
                xmlns:xsl="http://www.w3.org/1999/XSL/Transform">
<xsl:output method="text"/>
<xsl:strip-space elements="*"/>

<xsl:template match="@* | node()">
    <xsl:copy>
        <xsl:apply-templates select="@* | node()[not(@xml:lang)] | node()[@xml:lang = en]" />
    </xsl:copy>
</xsl:template>

<xsl:template match="comment()"/>

<xsl:template match="p">
<xsl:text>&#xA;</xsl:text>
<xsl:apply-templates />
<xsl:text>&#xA;</xsl:text>
</xsl:template>

<xsl:template match="ul">
<xsl:text>&#xA;</xsl:text>
<xsl:for-each select="li[not(@xml:lang)] | li[@xml:lang = en]">
<xsl:text>- </xsl:text>
<xsl:apply-templates />
<xsl:text>&#xA;</xsl:text>
</xsl:for-each>
</xsl:template>

<xsl:template match="ol">
<xsl:text>&#xA;</xsl:text>
<xsl:for-each select="li[not(@xml:lang)] | li[@xml:lang = en]">
<xsl:number count="li[not(@xml:lang)] | li[@xml:lang = en]" format="1. "/>
<xsl:apply-templates />
<xsl:text>&#xA;</xsl:text>
</xsl:for-each>
</xsl:template>

<xsl:template match="em">
<xsl:text>_</xsl:text>
<xsl:apply-templates />
<xsl:text>_</xsl:text>
</xsl:template>

<xsl:template match="code">
<xsl:apply-templates />
</xsl:template>

</xsl:stylesheet>
"""


def extract(relpath: str, *, workdir: str) -> ExtractedMetadata:
    if not relpath.endswith(".metainfo.xml") and not relpath.endswith(".appdata.xml"):
        raise _errors.UnhandledFileError(relpath, "appstream")

    dom = _get_transformed_dom(os.path.join(workdir, relpath))

    common_id = _get_value_from_xml_element(dom, "id")
    summary = _get_value_from_xml_element(dom, "summary")
    description = _get_value_from_xml_element(dom, "description")
    title = _get_value_from_xml_element(dom, "name")
    version = _get_latest_release_from_nodes(dom.findall("releases/release"))

    desktop_file_paths = []
    desktop_file_ids = _get_desktop_file_ids_from_nodes(dom.findall("launchable"))
    # if there are no launchables, use the appstream id to take into
    # account the legacy appstream definitions
    if common_id and not desktop_file_ids:
        if common_id.endswith(".desktop"):
            desktop_file_ids.append(common_id)
        else:
            desktop_file_ids.append(common_id + ".desktop")

    for desktop_file_id in desktop_file_ids:
        desktop_file_path = _desktop_file_id_to_path(desktop_file_id, workdir=workdir)
        if desktop_file_path:
            desktop_file_paths.append(desktop_file_path)

    icon = _extract_icon(dom, workdir, desktop_file_paths)

    return ExtractedMetadata(
        common_id=common_id,
        title=title,
        summary=summary,
        description=description,
        version=version,
        icon=icon,
        desktop_file_paths=desktop_file_paths,
    )


def _get_transformed_dom(path: str):
    dom = _get_dom(path)
    transform = _get_xslt()
    return transform(dom)


def _get_dom(path: str) -> lxml.etree.ElementTree:
    try:
        return lxml.etree.parse(path)
    except lxml.etree.ParseError as e:
        raise _errors.AppstreamFileParseError(path) from e


def _get_xslt():
    global _xslt
    if _xslt is not None:
        return _xslt

    xslt = lxml.etree.parse(StringIO(_XSLT))
    return lxml.etree.XSLT(xslt)


def _get_value_from_xml_element(tree, key) -> Optional[str]:
    node = tree.find(key)
    if node is not None and node.text:
        # Lines that should be empty end up with empty space after the
        # transformation. One example of this is seen for paragraphs (i.e.; <p>)
        # than hold list in then (i.e.; <ol> or <ul>) so we split all lines
        # here and strip any unwanted space.
        # TODO: Improve the XSLT to remove the need for this.
        return "\n".join([n.strip() for n in node.text.splitlines()]).strip()
    else:
        return None


def _get_latest_release_from_nodes(nodes) -> Optional[str]:
    for node in nodes:
        if "version" in node.attrib:
            return node.attrib["version"]
    return None


def _get_desktop_file_ids_from_nodes(nodes) -> List[str]:
    desktop_file_ids = []  # type: List[str]
    for node in nodes:
        if "type" in node.attrib and node.attrib["type"] == "desktop-id":
            desktop_file_ids.append(node.text.strip())
    return desktop_file_ids


def _desktop_file_id_to_path(desktop_file_id: str, *, workdir: str) -> Optional[str]:
    # For details about desktop file ids and their corresponding paths, see
    # https://standards.freedesktop.org/desktop-entry-spec/desktop-entry-spec-latest.html#desktop-file-id  # noqa
    for xdg_data_dir in ("usr/local/share", "usr/share"):
        desktop_file_path = os.path.join(
            xdg_data_dir, "applications", desktop_file_id.replace("-", "/")
        )
        # Check if it exists in workdir, but do not add it to the resulting path
        # as it later needs to exist in the prime directory to effectively be
        # used.
        if os.path.exists(os.path.join(workdir, desktop_file_path)):
            return desktop_file_path
    else:
        return None


def _extract_icon(dom, workdir: str, desktop_file_paths: List[str]) -> Optional[str]:
    icon_node = dom.find("icon")
    if icon_node is not None and "type" in icon_node.attrib:
        icon_node_type = icon_node.attrib["type"]
    else:
        icon_node_type = None

    icon = icon_node.text.strip() if icon_node is not None else None

    if icon_node_type == "remote":
        return icon
    elif icon_node_type == "stock":
        return _get_icon_from_theme(workdir, "hicolor", icon)

    # If an icon path is specified and the icon file exists, we'll use that, otherwise
    # we'll fall back to what's listed in the desktop file.
    if icon is None:
        return _get_icon_from_desktop_file(workdir, desktop_file_paths)
    elif os.path.exists(os.path.join(workdir, icon.lstrip("/"))):
        return icon
    else:
        return _get_icon_from_desktop_file(workdir, desktop_file_paths)


def _get_icon_from_desktop_file(
    workdir: str, desktop_file_paths: List[str]
) -> Optional[str]:
    # Icons in the desktop file can be either a full path to the icon file, or a name
    # to be searched in the standard locations. If the path is specified, use that,
    # otherwise look for the icon in the hicolor theme (also covers icon type="stock").
    # See https://standards.freedesktop.org/icon-theme-spec/icon-theme-spec-latest.html
    # for further information.
    for path in desktop_file_paths:
        entry = DesktopEntry()
        entry.parse(os.path.join(workdir, path))
        icon = entry.getIcon()
        icon_path = (
            icon
            if os.path.isabs(icon)
            else _get_icon_from_theme(workdir, "hicolor", icon)
        )
        return icon_path

    return None


def _get_icon_from_theme(workdir: str, theme: str, icon: str) -> Optional[str]:
    # Icon themes can carry icons in different pre-rendered sizes or scalable. Scalable
    # implementation is optional, so we'll try the largest pixmap and then scalable if
    # no other sizes are available.
    theme_dir = os.path.join("usr", "share", "icons", theme)
    if not os.path.exists(os.path.join(workdir, theme_dir)):
        return None

    # TODO: use index.theme
    entries = os.listdir(os.path.join(workdir, theme_dir))
    # size is NxN
    x_entries = (e.split("x") for e in entries if "x" in e)
    sized_entries = (e[0] for e in x_entries if e[0] == e[1])
    sizes = {}
    for icon_size_entry in sized_entries:
        with contextlib.suppress(ValueError):
            sizes[int(icon_size_entry)] = "{0}x{0}".format(icon_size_entry)

    icon_size = None
    if sizes:
        size = max(sizes.items(), key=operator.itemgetter(1))[0]
        icon_size = sizes[size]
        suffixes = [".png", ".xpm"]
    elif "scalable" in entries:
        icon_size = "scalable"
        suffixes = [".svg", ".svgz"]

    icon_path = None
    if icon_size:
        for suffix in suffixes:
            icon_path = os.path.join(theme_dir, icon_size, "apps", icon + suffix)
            if os.path.exists(os.path.join(workdir, icon_path)):
                break

    return icon_path



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018, 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import pathlib
from unittest import mock

import pytest

from snapcraft.internal import errors
from snapcraft.internal.meta.snap import Snap
from snapcraft.plugins.v1 import gradle
from snapcraft.project import Project

from . import PluginsV1BaseTestCase


def test_schema():
    assert gradle.GradlePlugin.schema() == {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "additionalProperties": False,
        "properties": {
            "gradle-openjdk-version": {"default": "", "type": "string"},
            "gradle-options": {
                "default": [],
                "items": {"type": "string"},
                "minitems": 1,
                "type": "array",
                "uniqueItems": True,
            },
            "gradle-output-dir": {"default": "build/libs", "type": "string"},
            "gradle-version": {"type": "string"},
            "gradle-version-checksum": {"type": "string"},
        },
        "required": ["source"],
        "type": "object",
    }


def test_pull_properties():
    assert gradle.GradlePlugin.get_pull_properties() == [
        "gradle-version",
        "gradle-version-checksum",
        "gradle-openjdk-version",
    ]


def test_build_properties():
    assert gradle.GradlePlugin.get_build_properties() == [
        "gradle-options",
        "gradle-output-dir",
    ]


def _get_expected_java_version(gradle_plugin) -> str:
    base = gradle_plugin.project._snap_meta.base
    ant_openjdk_version = gradle_plugin.options.gradle_openjdk_version

    if ant_openjdk_version:
        expected_java_version = ant_openjdk_version
    elif base in ("core", "core16"):
        expected_java_version = "9"
    else:
        expected_java_version = "11"

    return expected_java_version


_BASE_JAVA_COMBINATIONS = [
    ("", "core"),
    ("8", "core"),
    ("", "core16"),
    ("8", "core16"),
    ("", "core18"),
    ("11", "core18"),
]


@pytest.fixture(params=_BASE_JAVA_COMBINATIONS)
def gradle_plugin(tmp_path, request):
    """Return an instance of GradlePlugin setup with different bases and java versions."""
    java_version, base = request.param

    class Options:
        gradle_options = []
        gradle_output_dir = "build/libs"
        gradle_version = gradle._DEFAULT_GRADLE_VERSION
        gradle_version_checksum = gradle._DEFAULT_GRADLE_CHECKSUM
        gradle_openjdk_version = java_version

    os.chdir(tmp_path)
    project = Project()
    project._snap_meta = Snap(name="test-snap", base=base, confinement="strict")

    return gradle.GradlePlugin("test-part", Options(), project)


@pytest.fixture()
def gradle_plugin_with_assets(gradle_plugin, mock_zip):
    """Return an instance of GradlePlugin with artifacts setup using gradle_plugin."""
    pathlib.Path(gradle_plugin.sourcedir).mkdir(parents=True)

    expected_java_version = _get_expected_java_version(gradle_plugin)

    fake_java_path = (
        pathlib.Path(gradle_plugin.installdir)
        / f"usr/lib/jvm/java-{expected_java_version}-openjdk-amd64/bin/java"
    )
    fake_java_path.parent.mkdir(parents=True)
    fake_java_path.touch()

    gradle_tar_path = (
        pathlib.Path(gradle_plugin.partdir)
        / f"gradle/apache-gradle-{gradle_plugin.options.gradle_version}-bin.zip"
    )
    gradle_tar_path.parent.mkdir(parents=True)
    gradle_tar_path.touch()

    return gradle_plugin


def test_stage_and_build_packages(gradle_plugin):
    expected_java_version = _get_expected_java_version(gradle_plugin)

    assert gradle_plugin.stage_packages == [
        f"openjdk-{expected_java_version}-jre-headless"
    ]
    assert gradle_plugin.build_packages == [
        f"openjdk-{expected_java_version}-jdk-headless",
        "ca-certificates-java",
    ]


def test_build_gradlew(mock_run, gradle_plugin_with_assets):
    plugin = gradle_plugin_with_assets
    (pathlib.Path(plugin.sourcedir) / "gradlew").touch()

    def fake_run(l, **kwargs):
        os.makedirs(os.path.join(plugin.builddir, "build", "libs"))
        open(os.path.join(plugin.builddir, "build", "libs", "dummy.jar"), "w").close()

    mock_run.side_effect = fake_run

    plugin.build()

    mock_run.assert_called_once_with(
        ["./gradlew", "jar"], cwd=plugin.builddir, env=None
    )


def test_build_gradle(mock_run, gradle_plugin_with_assets):
    plugin = gradle_plugin_with_assets

    def fake_run(l, **kwargs):
        os.makedirs(os.path.join(plugin.builddir, "build", "libs"))
        open(os.path.join(plugin.builddir, "build", "libs", "dummy.jar"), "w").close()

    mock_run.side_effect = fake_run

    plugin.build()

    mock_run.assert_called_once_with(
        ["gradle", "jar"], cwd=plugin.builddir, env=mock.ANY
    )


def test_build_war_gradle(mock_run, gradle_plugin_with_assets):
    plugin = gradle_plugin_with_assets

    def fake_run(l, **kwargs):
        os.makedirs(os.path.join(plugin.builddir, "build", "libs"))
        open(os.path.join(plugin.builddir, "build", "libs", "dummy.war"), "w").close()

    mock_run.side_effect = fake_run

    plugin.build()

    mock_run.assert_called_once_with(
        ["gradle", "jar"], cwd=plugin.builddir, env=mock.ANY
    )


def test_build_war_gradlew(mock_run, gradle_plugin_with_assets):
    plugin = gradle_plugin_with_assets
    (pathlib.Path(plugin.sourcedir) / "gradlew").touch()

    def fake_run(l, **kwargs):
        os.makedirs(os.path.join(plugin.builddir, "build", "libs"))
        open(os.path.join(plugin.builddir, "build", "libs", "dummy.war"), "w").close()

    mock_run.side_effect = fake_run

    plugin.build()

    mock_run.assert_called_once_with(
        ["./gradlew", "jar"], cwd=plugin.builddir, env=None
    )


class TestGradleProxies:

    scenarios = [
        (
            "http proxy url",
            dict(
                env_var=("http_proxy", "http://test_proxy"),
                expected_args=["-Dhttp.proxyHost=test_proxy"],
            ),
        ),
        (
            "http proxy url and port",
            dict(
                env_var=("http_proxy", "http://test_proxy:3000"),
                expected_args=["-Dhttp.proxyHost=test_proxy", "-Dhttp.proxyPort=3000"],
            ),
        ),
        (
            "authenticated http proxy url",
            dict(
                env_var=("http_proxy", "http://user:pass@test_proxy:3000"),
                expected_args=[
                    "-Dhttp.proxyHost=test_proxy",
                    "-Dhttp.proxyPort=3000",
                    "-Dhttp.proxyUser=user",
                    "-Dhttp.proxyPassword=pass",
                ],
            ),
        ),
        (
            "https proxy url",
            dict(
                env_var=("https_proxy", "https://test_proxy"),
                expected_args=["-Dhttps.proxyHost=test_proxy"],
            ),
        ),
        (
            "https proxy url and port",
            dict(
                env_var=("https_proxy", "https://test_proxy:3000"),
                expected_args=[
                    "-Dhttps.proxyHost=test_proxy",
                    "-Dhttps.proxyPort=3000",
                ],
            ),
        ),
        (
            "authenticated https proxy url",
            dict(
                env_var=("https_proxy", "http://user:pass@test_proxy:3000"),
                expected_args=[
                    "-Dhttps.proxyHost=test_proxy",
                    "-Dhttps.proxyPort=3000",
                    "-Dhttps.proxyUser=user",
                    "-Dhttps.proxyPassword=pass",
                ],
            ),
        ),
    ]

    def test_build_gradlew(
        self, monkeypatch, mock_run, gradle_plugin_with_assets, env_var, expected_args
    ):
        monkeypatch.setenv(*env_var)

        plugin = gradle_plugin_with_assets
        (pathlib.Path(plugin.sourcedir) / "gradlew").touch()

        def fake_run(l, **kwargs):
            os.makedirs(os.path.join(plugin.builddir, "build", "libs"))
            open(
                os.path.join(plugin.builddir, "build", "libs", "dummy.jar"), "w"
            ).close()

        mock_run.side_effect = fake_run

        plugin.build()

        mock_run.assert_called_once_with(
            ["./gradlew"] + expected_args + ["jar"], cwd=plugin.builddir, env=None
        )

    def test_build_gradle(
        self, monkeypatch, mock_run, gradle_plugin_with_assets, env_var, expected_args
    ):
        monkeypatch.setenv(*env_var)

        plugin = gradle_plugin_with_assets

        def fake_run(l, **kwargs):
            os.makedirs(os.path.join(plugin.builddir, "build", "libs"))
            open(
                os.path.join(plugin.builddir, "build", "libs", "dummy.jar"), "w"
            ).close()

        mock_run.side_effect = fake_run

        plugin.build()

        mock_run.assert_called_once_with(
            ["gradle"] + expected_args + ["jar"], cwd=plugin.builddir, env=mock.ANY
        )


class GradlePluginUnsupportedBase(PluginsV1BaseTestCase):
    def setUp(self):
        super().setUp()

        self.project._snap_meta.base = "unsupported-base"

        class Options:
            gradle_options = []
            gradle_output_dir = "build/libs"
            gradle_version = gradle._DEFAULT_GRADLE_VERSION
            gradle_version_checksum = gradle._DEFAULT_GRADLE_CHECKSUM
            gradle_openjdk_version = ""

        self.options = Options()

    def test_unsupported_base_raises(self):
        self.assertRaises(
            errors.PluginBaseError,
            gradle.GradlePlugin,
            "test-part",
            self.options,
            self.project,
        )


class TestUnsupportedJDKVersionError:
    scenarios = (
        (
            "core",
            dict(
                base="core",
                version="11",
                expected_message=(
                    "The gradle-openjdk-version plugin property was set to '11'.\n"
                    "Valid values for the 'core' base are: '8' or '9'."
                ),
            ),
        ),
        (
            "core16",
            dict(
                base="core16",
                version="11",
                expected_message=(
                    "The gradle-openjdk-version plugin property was set to '11'.\n"
                    "Valid values for the 'core16' base are: '8' or '9'."
                ),
            ),
        ),
        (
            "core18",
            dict(
                base="core18",
                version="9",
                expected_message=(
                    "The gradle-openjdk-version plugin property was set to '9'.\n"
                    "Valid values for the 'core18' base are: '11' or '8'."
                ),
            ),
        ),
    )

    def test_use_invalid_openjdk_version_fails(self, base, version, expected_message):
        class Options:
            gradle_options = []
            gradle_output_dir = "build/libs"
            gradle_version = gradle._DEFAULT_GRADLE_VERSION
            gradle_version_checksum = gradle._DEFAULT_GRADLE_CHECKSUM
            gradle_openjdk_version = version

        project = Project()
        project._snap_meta = Snap(name="test-snap", base=base, confinement="strict")

        with pytest.raises(gradle.UnsupportedJDKVersionError) as error:
            gradle.GradlePlugin("test-part", Options(), project)
            assert str(error) == expected_message



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from unittest import mock
from unittest.mock import call

import launchpadlib
import pytest

from snapcraft.internal.repo import apt_ppa, errors


@pytest.fixture
def mock_launchpad(autouse=True):
    with mock.patch(
        "snapcraft.internal.repo.apt_ppa.Launchpad",
        spec=launchpadlib.launchpad.Launchpad,
    ) as m:
        m.login_anonymously.return_value.load.return_value.signing_key_fingerprint = (
            "FAKE-PPA-SIGNING-KEY"
        )
        yield m


def test_split_ppa_parts():
    owner, name = apt_ppa.split_ppa_parts(ppa="test-owner/test-name")

    assert owner == "test-owner"
    assert name == "test-name"


def test_split_ppa_parts_invalid():
    with pytest.raises(errors.AptPPAInstallError) as exc_info:
        apt_ppa.split_ppa_parts(ppa="ppa-missing-slash")

    assert exc_info.value._ppa == "ppa-missing-slash"


def test_get_launchpad_ppa_key_id(mock_launchpad,):
    key_id = apt_ppa.get_launchpad_ppa_key_id(ppa="ppa-owner/ppa-name")

    assert key_id == "FAKE-PPA-SIGNING-KEY"
    assert mock_launchpad.mock_calls == [
        call.login_anonymously("snapcraft", "production"),
        call.login_anonymously().load("~ppa-owner/+archive/ppa-name"),
    ]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pytest

from snapcraft.internal import cache


@pytest.fixture()
def random_data_file(tmp_path):
    """Return a file with some text in it."""
    file_path = tmp_path / "to_hash"
    with file_path.open("w") as to_hash_file:
        print("random stub data", file=to_hash_file, end="")

    return file_path.as_posix()


@pytest.fixture()
def file_cache(xdg_dirs):
    """Return a FileCache instance."""
    return cache.FileCache()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from textwrap import dedent

from snapcraft.internal.review_tools import errors


class TestSnapcraftException:

    scenarios = (
        (
            "ReviewError (linting error with link)",
            {
                "exception_class": errors.ReviewError,
                "kwargs": dict(
                    review_json={
                        "snap.v2_functional": {"error": {}, "warn": {}},
                        "snap.v2_lint": {
                            "error": {
                                "lint-snap-v2:lint_issue": {
                                    "link": "https://issue-link",
                                    "text": "(NEEDS REVIEW) linting message.",
                                }
                            },
                            "warn": {},
                        },
                        "snap.v2_security": {"error": {}, "warn": {}},
                    }
                ),
                "expected_details": dedent(
                    """\
                    Linting Issues:
                    - (NEEDS REVIEW) linting message. (Refer to https://issue-link)"""
                ),
            },
        ),
        (
            "ReviewError (linting warning with link)",
            {
                "exception_class": errors.ReviewError,
                "kwargs": dict(
                    review_json={
                        "snap.v2_functional": {"error": {}, "warn": {}},
                        "snap.v2_lint": {
                            "warn": {
                                "lint-snap-v2:lint_issue": {
                                    "link": "https://issue-link",
                                    "text": "(NEEDS REVIEW) linting message.",
                                }
                            },
                            "error": {},
                        },
                        "snap.v2_security": {"error": {}, "warn": {}},
                    }
                ),
                "expected_details": dedent(
                    """\
                    Linting Issues:
                    - (NEEDS REVIEW) linting message. (Refer to https://issue-link)"""
                ),
            },
        ),
        (
            "ReviewError (linting error without link)",
            {
                "exception_class": errors.ReviewError,
                "kwargs": dict(
                    review_json={
                        "snap.v2_functional": {"error": {}, "warn": {}},
                        "snap.v2_lint": {
                            "error": {
                                "lint-snap-v2:lint_issue": {
                                    "text": "(NEEDS REVIEW) linting message."
                                }
                            },
                            "warn": {},
                        },
                        "snap.v2_security": {"error": {}, "warn": {}},
                    }
                ),
                "expected_details": dedent(
                    """\
                    Linting Issues:
                    - (NEEDS REVIEW) linting message."""
                ),
            },
        ),
        (
            "ReviewError (linting warning without link)",
            {
                "exception_class": errors.ReviewError,
                "kwargs": dict(
                    review_json={
                        "snap.v2_functional": {"error": {}, "warn": {}},
                        "snap.v2_lint": {
                            "warn": {
                                "lint-snap-v2:lint_issue": {
                                    "text": "(NEEDS REVIEW) linting message."
                                }
                            },
                            "error": {},
                        },
                        "snap.v2_security": {"error": {}, "warn": {}},
                    }
                ),
                "expected_details": dedent(
                    """\
                    Linting Issues:
                    - (NEEDS REVIEW) linting message."""
                ),
            },
        ),
        (
            "ReviewError (security error with link)",
            {
                "exception_class": errors.ReviewError,
                "kwargs": dict(
                    review_json={
                        "snap.v2_functional": {"error": {}, "warn": {}},
                        "snap.v2_security": {
                            "error": {
                                "security-snap-v2:security_issue": {
                                    "link": "https://issue-link",
                                    "text": "(NEEDS REVIEW) security message.",
                                }
                            },
                            "warn": {},
                        },
                        "snap.v2_lint": {"error": {}, "warn": {}},
                    }
                ),
                "expected_details": dedent(
                    """\
                    Security Issues:
                    - (NEEDS REVIEW) security message. (Refer to https://issue-link)"""
                ),
            },
        ),
        (
            "ReviewError (security warning with link)",
            {
                "exception_class": errors.ReviewError,
                "kwargs": dict(
                    review_json={
                        "snap.v2_functional": {"error": {}, "warn": {}},
                        "snap.v2_security": {
                            "warn": {
                                "security-snap-v2:security_issue": {
                                    "link": "https://issue-link",
                                    "text": "(NEEDS REVIEW) security message.",
                                }
                            },
                            "error": {},
                        },
                        "snap.v2_lint": {"error": {}, "warn": {}},
                    }
                ),
                "expected_details": dedent(
                    """\
                    Security Issues:
                    - (NEEDS REVIEW) security message. (Refer to https://issue-link)"""
                ),
            },
        ),
        (
            "ReviewError (security error without link)",
            {
                "exception_class": errors.ReviewError,
                "kwargs": dict(
                    review_json={
                        "snap.v2_functional": {"error": {}, "warn": {}},
                        "snap.v2_security": {
                            "error": {
                                "security-snap-v2:security_issue": {
                                    "text": "(NEEDS REVIEW) security message."
                                }
                            },
                            "warn": {},
                        },
                        "snap.v2_lint": {"error": {}, "warn": {}},
                    }
                ),
                "expected_details": dedent(
                    """\
                    Security Issues:
                    - (NEEDS REVIEW) security message."""
                ),
            },
        ),
        (
            "ReviewError (security warning without link)",
            {
                "exception_class": errors.ReviewError,
                "kwargs": dict(
                    review_json={
                        "snap.v2_functional": {"error": {}, "warn": {}},
                        "snap.v2_security": {
                            "warn": {
                                "security-snap-v2:security_issue": {
                                    "text": "(NEEDS REVIEW) security message."
                                }
                            },
                            "error": {},
                        },
                        "snap.v2_lint": {"error": {}, "warn": {}},
                    }
                ),
                "expected_details": dedent(
                    """\
                    Security Issues:
                    - (NEEDS REVIEW) security message."""
                ),
            },
        ),
        (
            "ReviewError (functional error with link)",
            {
                "exception_class": errors.ReviewError,
                "kwargs": dict(
                    review_json={
                        "snap.v2_security": {"error": {}, "warn": {}},
                        "snap.v2_functional": {
                            "error": {
                                "functional-snap-v2:functional_issue": {
                                    "link": "https://issue-link",
                                    "text": "(NEEDS REVIEW) functional message.",
                                }
                            },
                            "warn": {},
                        },
                        "snap.v2_lint": {"error": {}, "warn": {}},
                    }
                ),
                "expected_details": dedent(
                    """\
                    Functional Issues:
                    - (NEEDS REVIEW) functional message. (Refer to https://issue-link)"""
                ),
            },
        ),
        (
            "ReviewError (functional warning with link)",
            {
                "exception_class": errors.ReviewError,
                "kwargs": dict(
                    review_json={
                        "snap.v2_security": {"error": {}, "warn": {}},
                        "snap.v2_functional": {
                            "warn": {
                                "functional-snap-v2:functional_issue": {
                                    "link": "https://issue-link",
                                    "text": "(NEEDS REVIEW) functional message.",
                                }
                            },
                            "error": {},
                        },
                        "snap.v2_lint": {"error": {}, "warn": {}},
                    }
                ),
                "expected_details": dedent(
                    """\
                    Functional Issues:
                    - (NEEDS REVIEW) functional message. (Refer to https://issue-link)"""
                ),
            },
        ),
        (
            "ReviewError (functional error without link)",
            {
                "exception_class": errors.ReviewError,
                "kwargs": dict(
                    review_json={
                        "snap.v2_security": {"error": {}, "warn": {}},
                        "snap.v2_functional": {
                            "error": {
                                "functional-snap-v2:functional_issue": {
                                    "text": "(NEEDS REVIEW) functional message."
                                }
                            },
                            "warn": {},
                        },
                        "snap.v2_lint": {"error": {}, "warn": {}},
                    }
                ),
                "expected_details": dedent(
                    """\
                    Functional Issues:
                    - (NEEDS REVIEW) functional message."""
                ),
            },
        ),
        (
            "ReviewError (functional warning without link)",
            {
                "exception_class": errors.ReviewError,
                "kwargs": dict(
                    review_json={
                        "snap.v2_security": {"error": {}, "warn": {}},
                        "snap.v2_functional": {
                            "warn": {
                                "functional-snap-v2:functional_issue": {
                                    "text": "(NEEDS REVIEW) functional message."
                                }
                            },
                            "error": {},
                        },
                        "snap.v2_lint": {"error": {}, "warn": {}},
                    }
                ),
                "expected_details": dedent(
                    """\
                    Functional Issues:
                    - (NEEDS REVIEW) functional message."""
                ),
            },
        ),
        (
            "ReviewError",
            {
                "exception_class": errors.ReviewError,
                "kwargs": dict(
                    review_json={
                        "snap.v2_security": {
                            "error": {},
                            "warn": {
                                "functional-snap-v2:security_issue": {
                                    "text": "(NEEDS REVIEW) security message."
                                }
                            },
                        },
                        "snap.v2_functional": {
                            "warn": {
                                "functional-snap-v2:functional_issue": {
                                    "text": "(NEEDS REVIEW) functional message."
                                }
                            },
                            "error": {},
                        },
                        "snap.v2_lint": {
                            "error": {},
                            "warn": {
                                "functional-snap-v2:lint_issue": {
                                    "text": "(NEEDS REVIEW) lint message."
                                }
                            },
                        },
                    }
                ),
                "expected_details": dedent(
                    """\
                    Functional Issues:
                    - (NEEDS REVIEW) functional message.

                    Linting Issues:
                    - (NEEDS REVIEW) lint message.

                    Security Issues:
                    - (NEEDS REVIEW) security message."""
                ),
            },
        ),
    )

    def test_snapcraft_exception_handling(
        self, exception_class, expected_details, kwargs
    ):
        exception = exception_class(**kwargs)

        assert exception.get_details() == expected_details
        assert exception.get_brief() == "Review Tools did not fully pass for this snap."
        assert (
            exception.get_resolution()
            == "Specific measures might need to be taken on the Snap Store before this snap can be fully accepted."
        )
        assert exception.get_docs_url() is None
        assert exception.get_reportable() is False



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from unittest import mock

import pytest

from snapcraft.cli import lifecycle


@pytest.mark.parametrize(
    "output,pack_name,pack_dir",
    [
        ("/tmp/output.snap", "output.snap", "/tmp"),
        ("/tmp", None, "/tmp"),
        ("output.snap", "output.snap", None),
    ],
)
@pytest.mark.parametrize(
    "compression", ["xz", "lzo", None],
)
@mock.patch("snapcraft.file_utils.get_host_tool_path", return_value="/bin/snap")
@mock.patch("snapcraft.cli.lifecycle._run_pack", return_value="ignore.snap")
def test_pack(mock_run_pack, mock_host_tool, compression, output, pack_name, pack_dir):
    lifecycle._pack(directory="/my/snap", compression=compression, output=output)

    assert mock_host_tool.mock_calls == [
        mock.call(command_name="snap", package_name="snapd")
    ]

    pack_command = ["/bin/snap", "pack"]

    if compression:
        pack_command.extend(["--compression", compression])

    if pack_name:
        pack_command.extend(["--filename", pack_name])

    pack_command.append("/my/snap")

    if pack_dir:
        pack_command.append(pack_dir)

    assert mock_run_pack.mock_calls == [mock.call(pack_command)]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

import fixtures
from testtools.matchers import Equals

from snapcraft import ProjectOptions
from snapcraft import __version__ as snapcraft_version
from snapcraft.storeapi.http_clients import agent
from tests import unit
from tests.fixture_setup.os_release import FakeOsRelease


class UserAgentTestCase(unit.TestCase):
    def test_user_agent_linux(self):
        self.useFixture(FakeOsRelease())

        arch = ProjectOptions().deb_arch
        expected = f"snapcraft/{snapcraft_version} Ubuntu/16.04 ({arch})"

        self.expectThat(agent.get_user_agent("linux"), Equals(expected))

    def test_user_agent_linux_unknown(self):
        self.useFixture(FakeOsRelease(name=None, version_id=None))

        arch = ProjectOptions().deb_arch
        expected = f"snapcraft/{snapcraft_version} Unknown/Unknown Version ({arch})"

        self.expectThat(agent.get_user_agent("linux"), Equals(expected))

    def test_user_agent_windows(self):
        arch = ProjectOptions().deb_arch
        expected = f"snapcraft/{snapcraft_version} Windows ({arch})"

        self.expectThat(agent.get_user_agent(platform="windows"), Equals(expected))

    def test_user_agent_darwin(self):
        arch = ProjectOptions().deb_arch
        expected = f"snapcraft/{snapcraft_version} Darwin ({arch})"

        self.expectThat(agent.get_user_agent(platform="darwin"), Equals(expected))

    def test_in_travis_ci_env(self):
        self.useFixture(fixtures.EnvironmentVariable("TRAVIS_TESTING", "1"))

        self.assertTrue(agent._is_ci_env())

    def test_in_autopkgtest_ci_env(self):
        self.useFixture(fixtures.EnvironmentVariable("AUTOPKGTEST_TMP", "1"))

        self.assertTrue(agent._is_ci_env())

    def test_not_in_ci_env(self):
        # unset any known testing environment vars
        testing_vars = ["TRAVIS", "AUTHPKGTEST_TMP"]
        vars_to_unset = []
        for env_var in os.environ:
            for test_var in testing_vars:
                if env_var.startswith(test_var):
                    vars_to_unset.append(env_var)

        for var in vars_to_unset:
            self.useFixture(fixtures.EnvironmentVariable(var, None))

        self.assertFalse(agent._is_ci_env())



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The npm plugin is useful for node based parts that use npm.

The plugin uses npm to install dependencies from `package.json`. It
also sets up binaries defined in `package.json` into the `PATH`.

This plugin uses the common plugin keywords as well as those for "sources".
For more information check the 'plugins' topic for the former and the
'sources' topic for the latter.

Additionally, this plugin uses the following plugin-specific keywords:

    - npm-node-version
      (string)
      The version of nodejs you want the snap to run on.
      This includes npm, as would be downloaded from https://nodejs.org
"""

import os
import platform
from textwrap import dedent
from typing import Any, Dict, List, Set

from snapcraft.plugins.v2 import PluginV2

_NODE_ARCH_FROM_SNAP_ARCH = {
    "i386": "x86",
    "amd64": "x64",
    "armhf": "armv7l",
    "arm64": "arm64",
    "ppc64el": "ppc64le",
    "s390x": "s390x",
}
_NODE_ARCH_FROM_PLATFORM = {"x86_64": {"32bit": "x86", "64bit": "x64"}}


class NpmPlugin(PluginV2):
    @classmethod
    def get_schema(cls) -> Dict[str, Any]:
        return {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "type": "object",
            "additionalProperties": False,
            "properties": {"npm-node-version": {"type": "string"}},
            "required": ["source", "npm-node-version"],
        }

    @staticmethod
    def _get_architecture() -> str:
        snap_arch = os.getenv("SNAP_ARCH")
        # The first scenario is the general case as snapcraft will be running from the snap.
        if snap_arch is not None:
            node_arch = _NODE_ARCH_FROM_SNAP_ARCH[snap_arch]
        # But there may be times when running from a virtualenv while doing development.
        else:
            node_arch = _NODE_ARCH_FROM_PLATFORM[platform.machine()][
                platform.architecture()[0]
            ]

        return node_arch

    def get_build_snaps(self) -> Set[str]:
        return set()

    def get_build_packages(self) -> Set[str]:
        return {"curl", "gcc"}

    def _get_node_command(self) -> str:
        arch = self._get_architecture()
        version = self.options.npm_node_version

        node_uri = (
            f"https://nodejs.org/dist/v{version}/node-v{version}-linux-{arch}.tar.gz"
        )
        # TODO snapcraftctl for downloading assets.
        return dedent(
            f"""\
        if [ ! -f "${{SNAPCRAFT_PART_INSTALL}}/bin/node" ]; then
            curl -s "{node_uri}" | tar xzf - -C "${{SNAPCRAFT_PART_INSTALL}}/" --strip-components=1
        fi
        """
        )

    def get_build_environment(self) -> Dict[str, str]:
        return dict(PATH="${SNAPCRAFT_PART_INSTALL}/bin:${PATH}")

    def get_build_commands(self) -> List[str]:
        return [
            self._get_node_command(),
            'npm install -g --prefix "${SNAPCRAFT_PART_INSTALL}" $(npm pack . | tail -1)',
        ]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pathlib
from textwrap import dedent

import pytest

from snapcraft.internal import project_loader
from snapcraft.project import Project


def get_project_config(snapcraft_yaml_content):
    snapcraft_yaml_path = pathlib.Path("snapcraft.yaml")
    with snapcraft_yaml_path.open("w") as snapcraft_yaml_file:
        print(snapcraft_yaml_content, file=snapcraft_yaml_file)

    project = Project(snapcraft_yaml_file_path=snapcraft_yaml_path.as_posix())
    return project_loader.load_config(project)


def test_build_packages_from_snapcraft_yaml(tmp_work_path):
    snapcraft_yaml = dedent(
        """\
        name: test
        base: core18
        version: "1.0"
        summary: test
        description: test
        confinement: strict
        grade: stable

        build-packages: [foobar]

        parts:
          part1:
            plugin: nil
            build-packages: [foo, bar]
        """
    )

    project_config = get_project_config(snapcraft_yaml)

    assert project_config.get_build_packages() == {"foobar", "foo", "bar"}


@pytest.mark.parametrize(
    "source,expected_package",
    [("git://github.com/ubuntu-core/snapcraft.git", "git"), ("lp:ubuntu-push", "bzr")],
)
def test_config_adds_vcs_packages_to_build_packages(
    tmp_work_path, source, expected_package
):
    snapcraft_yaml = dedent(
        f"""\
            name: test
            base: core18
            version: "1"
            summary: test
            description: test
            confinement: strict
            grade: stable

            parts:
              part1:
                source: {source}
                plugin: nil
            """
    )

    project_config = get_project_config(snapcraft_yaml)

    assert expected_package in project_config.get_build_packages()


@pytest.mark.parametrize(
    "source_type,expected_package",
    [
        ("git", "git"),
        ("hg", "mercurial"),
        ("mercurial", "mercurial"),
        ("bzr", "bzr"),
        ("svn", "subversion"),
        ("subversion", "subversion"),
    ],
)
def test_config_adds_vcs_packages_to_build_packages_from_types(
    tmp_work_path, source_type, expected_package
):
    snapcraft_yaml = dedent(
        f"""\
            name: test
            base: core18
            version: "1"
            summary: test
            description: test
            confinement: strict
            grade: stable

            parts:
              part1:
                source: http://something/somewhere
                source-type: {source_type}
                plugin: autotools
        """
    )

    project_config = get_project_config(snapcraft_yaml)

    assert expected_package in project_config.get_build_packages()


def test_git_added_for_version_git(tmp_work_path):
    snapcraft_yaml = dedent(
        """\
            name: test
            base: core18
            version: "git"
            summary: test
            description: test
            confinement: strict
            grade: stable

            parts:
              part1:
                plugin: nil
            """
    )

    project_config = get_project_config(snapcraft_yaml)

    "git" in project_config.get_build_packages()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import textwrap
from typing import Tuple
from unittest import mock

from testtools.matchers import Equals

from snapcraft.internal.project_loader import errors, supported_extension_names
from snapcraft.internal.project_loader._extensions._extension import Extension
from tests import fixture_setup

from . import CommandBaseTestCase


class ExtensionsCommandTest(CommandBaseTestCase):
    def setUp(self):
        super().setUp()

        # Create a few fake extensions
        self.useFixture(_test1_extension_fixture())
        self.useFixture(_test2_extension_fixture())
        self.useFixture(_test3_extension_fixture())
        self.useFixture(_test4_extension_fixture())

    @mock.patch(
        "pkgutil.iter_modules",
        return_value=(
            (None, "test1", None),
            (None, "test2", None),
            (None, "test3", None),
            (None, "_test4", None),
        ),
    )
    def test_list_extensions(self, fake_iter_modules):
        result = self.run_command(["extensions"])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output,
            Equals(
                textwrap.dedent(
                    """\
                    Extension name    Supported bases
                    ----------------  -----------------
                    test1             core16
                    test2             core16
                    test3             core16, core18
                    """
                )
            ),
        )

    def test_extension(self):
        result = self.run_command(["extension", "test1"])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output,
            Equals(
                textwrap.dedent(
                    """\
                    This is the Test1 extension.

                    It does stuff.
                    """
                )
            ),
        )

        result = self.run_command(["extension", "test2"])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output,
            Equals(
                textwrap.dedent(
                    """\
                    This is the Test2 extension.

                    It does other stuff.
                    """
                )
            ),
        )

    def test_extension_missing_docs(self):
        raised = self.assertRaises(
            errors.ExtensionMissingDocumentationError,
            self.run_command,
            ["extension", "test3"],
        )

        self.assertThat(
            str(raised),
            Equals(
                "The 'test3' extension appears to be missing documentation.\n"
                "We would appreciate it if you created a bug report about this at "
                "https://launchpad.net/snapcraft/+filebug"
            ),
        )

    def test_expand_extensions(self):
        self.make_snapcraft_yaml(
            textwrap.dedent(
                """\
                name: test
                version: '1'
                summary: test
                description: test
                base: core16
                grade: stable
                confinement: strict

                apps:
                    test-app:
                        command: echo "hello"
                        extensions: [test1]

                parts:
                    test-part:
                        plugin: nil
                """
            )
        )

        result = self.run_command(["expand-extensions"])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output,
            Equals(
                textwrap.dedent(
                    """\
                    name: test
                    version: '1'
                    summary: test
                    description: test
                    base: core16
                    grade: stable
                    confinement: strict
                    apps:
                      test-app:
                        command: echo "hello"
                        environment:
                          EXTENSION_NAME: test1
                    parts:
                      test-part:
                        plugin: nil
                        after:
                        - extension-part
                      extension-part:
                        plugin: nil
                    """
                )
            ),
        )

    def test_all_extension_docs(self):
        for extension_name in supported_extension_names():
            try:
                self.run_command(["extension", extension_name])
            except Exception:
                self.fail(
                    "The {!r} extension appears to be missing documentation".format(
                        extension_name
                    )
                )


def _test1_extension_fixture():
    class ExtensionImpl(Extension):
        """This is the Test1 extension.

        It does stuff.
        """

        @staticmethod
        def get_supported_bases() -> Tuple[str, ...]:
            return ("core16",)

        @staticmethod
        def get_supported_confinement() -> Tuple[str, ...]:
            return ("strict",)

        def __init__(self, extension_name, yaml_data):
            super().__init__(extension_name=extension_name, yaml_data=yaml_data)
            self.app_snippet = {"environment": {"EXTENSION_NAME": "test1"}}
            self.part_snippet = {"after": ["extension-part"]}
            self.parts = {"extension-part": {"plugin": "nil"}}

    return fixture_setup.FakeExtension("test1", ExtensionImpl)


def _test2_extension_fixture():
    class ExtensionImpl(Extension):
        """This is the Test2 extension.

        It does other stuff.
        """

        @staticmethod
        def get_supported_bases() -> Tuple[str, ...]:
            return ("core16",)

        @staticmethod
        def get_supported_confinement() -> Tuple[str, ...]:
            return ("strict",)

        def __init__(self, extension_name, yaml_data):
            super().__init__(extension_name=extension_name, yaml_data=yaml_data)
            self.part_snippet = {"after": ["extension-part"]}
            self.parts = {"extension-part": {"plugin": "nil"}}

    return fixture_setup.FakeExtension("test2", ExtensionImpl)


def _test3_extension_fixture():
    class ExtensionImpl(Extension):
        @staticmethod
        def get_supported_bases() -> Tuple[str, ...]:
            return ("core16", "core18")

        @staticmethod
        def get_supported_confinement() -> Tuple[str, ...]:
            return ("strict",)

        def __init__(self, extension_name, yaml_data):
            super().__init__(extension_name=extension_name, yaml_data=yaml_data)
            self.parts = {"extension-part": {"plugin": "nil"}}

    return fixture_setup.FakeExtension("test3", ExtensionImpl)


def _test4_extension_fixture():
    class ExtensionImpl(Extension):
        @staticmethod
        def get_supported_bases() -> Tuple[str, ...]:
            return ("core16", "core18")

        @staticmethod
        def get_supported_confinement() -> Tuple[str, ...]:
            return ("strict",)

        def __init__(self, extension_name, yaml_data):
            super().__init__(extension_name=extension_name, yaml_data=yaml_data)
            self.parts = {"extension-part": {"plugin": "nil"}}

    return fixture_setup.FakeExtension("_test4", ExtensionImpl)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import pathlib
from typing import List
from unittest import mock

import pytest
import xdg


def pytest_generate_tests(metafunc):
    idlist = []
    argvalues = []
    if metafunc.cls is None:
        return

    for scenario in metafunc.cls.scenarios:
        idlist.append(scenario[0])
        items = scenario[1].items()
        argnames = [x[0] for x in items]
        argvalues.append([x[1] for x in items])
    metafunc.parametrize(argnames, argvalues, ids=idlist, scope="class")


@pytest.fixture
def mock_subprocess_run():
    """A no-op subprocess.run mock."""
    patcher = mock.patch("subprocess.run")
    yield patcher.start()
    patcher.stop()


@pytest.fixture
def tmp_work_path(tmp_path):
    """Setup a temporary directory and chdir to it."""
    os.chdir(tmp_path)
    return tmp_path


@pytest.fixture
def xdg_dirs(tmp_path, monkeypatch):
    """Setup XDG directories in a temporary directory."""
    monkeypatch.setattr(
        xdg.BaseDirectory, "xdg_config_home", (tmp_path / ".config").as_posix()
    )
    monkeypatch.setattr(
        xdg.BaseDirectory, "xdg_data_home", (tmp_path / ".local").as_posix()
    )
    monkeypatch.setattr(
        xdg.BaseDirectory, "xdg_cache_home", (tmp_path / ".cache").as_posix()
    )
    monkeypatch.setattr(
        xdg.BaseDirectory,
        "xdg_config_dirs",
        lambda: [(tmp_path / ".config").as_posix()],
    )
    monkeypatch.setattr(
        xdg.BaseDirectory, "xdg_data_dirs", lambda: [(tmp_path / ".config").as_posix()]
    )

    monkeypatch.setenv("XDG_CONFIG_HOME", (tmp_path / ".config").as_posix())
    monkeypatch.setenv("XDG_DATA_HOME", (tmp_path / ".local").as_posix())
    monkeypatch.setenv("XDG_CACHE_HOME", (tmp_path / ".cache").as_posix())

    return tmp_path


@pytest.fixture()
def in_snap(monkeypatch):
    """Simualte being run from within the context of the Snapcraft snap."""
    monkeypatch.setenv("SNAP", "/snap/snapcraft/current")
    monkeypatch.setenv("SNAP_NAME", "snapcraft")
    monkeypatch.setenv("SNAP_VERSION", "4.0")


@pytest.fixture()
def fake_exists(monkeypatch):
    """Fakely return True when checking for preconfigured paths."""

    class FileCheck:
        def __init__(self) -> None:
            self._original_exists = os.path.exists
            self.paths: List[str] = list()

        def exists(self, path: str) -> bool:
            if pathlib.Path(path) in self.paths:
                return True
            return self._original_exists(path)

    file_checker = FileCheck()
    monkeypatch.setattr(os.path, "exists", file_checker.exists)

    return file_checker



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess

from testtools.matchers import Equals, FileExists

from tests import integration


class SubversionSourceTestCase(integration.SubversionSourceBaseTestCase):
    def test_pull_svn_checkout(self):
        self.copy_project_to_cwd("svn-pull")

        self.init_source_control()
        self.checkout("file:///{}".format(os.path.join(self.path, "repo")), "local")
        open(os.path.join("local", "file"), "w").close()
        self.add("file", cwd="local/")
        self.commit("test", cwd="local/")
        self.update(cwd="local/")
        subprocess.check_call(["rm", "-rf", "local/"], stdout=subprocess.DEVNULL)

        self.run_snapcraft("pull")
        part_src_path = os.path.join(self.parts_dir, "svn", "src")
        revno = subprocess.check_output(["svnversion", part_src_path]).strip()
        self.assertThat(revno, Equals(b"1"))
        self.assertThat(os.path.join(part_src_path, "file"), FileExists())

    def test_pull_svn_update(self):
        self.copy_project_to_cwd("svn-pull-update")

        self.init_source_control()

        self.checkout("file:///{}".format(os.path.join(self.path, "repo")), "local")
        open(os.path.join("local", "file"), "w").close()
        self.add("file", cwd="local/")
        self.commit("test", cwd="local/")
        self.update(cwd="local/")
        subprocess.check_call(["rm", "-rf", "local/"], stdout=subprocess.DEVNULL)

        part_src_path = os.path.join(self.parts_dir, "svn", "src")
        self.checkout(
            "file:///{}".format(os.path.join(self.path, "repo")), part_src_path
        )
        self.checkout("file:///{}".format(os.path.join(self.path, "repo")), "local")
        open(os.path.join("local", "filetwo"), "w").close()
        self.add("filetwo", cwd="local/")
        self.commit("testtwo", cwd="local/")
        self.update(cwd="local/")
        subprocess.check_call(["rm", "-rf", "local/"], stdout=subprocess.DEVNULL)

        self.run_snapcraft("pull")
        revno = subprocess.check_output(["svnversion", part_src_path]).strip()
        self.assertThat(revno, Equals(b"2"))
        self.assertThat(os.path.join(part_src_path, "file"), FileExists())
        self.assertThat(os.path.join(part_src_path, "filetwo"), FileExists())
